from __future__ import annotations

from abc import abstractmethod


class ClientParametersBuilder:
    """
    This class is used to build the parameters for the client call.
    Depending on the client, the parameters are built differently.
    """

    @abstractmethod
    def build_parameters(self, category: str, operation: str, **kwargs) -> dict:
        """
        Build the parameters for the client call.
        """



from __future__ import annotations

from pydantic import BaseModel


class ClientConfig(BaseModel):
    """Client configuration model."""

    endpoint: str
    """API endpoint."""


class BasicAuth(ClientConfig):
    """Basic authentication model."""

    user: str
    """Username."""

    password: str
    """Basic authentication password."""


class OAuth2TokenAuth(ClientConfig):
    """OAuth2 token authentication model."""

    access_token: str
    """OAuth2 token."""

    refresh_token: str
    """OAuth2 refresh token."""

    client_id: str
    """OAuth2 client id."""

    issuer_endpoint: str
    """OAuth2 issuer endpoint."""



from __future__ import annotations

from typing import Any

from digitalhub.entities._base.entity._constructors.uuid import build_uuid
from digitalhub.entities._base.material.utils import build_log_path_from_source, eval_local_source
from digitalhub.entities._commons.enums import EntityTypes


def eval_source(
    source: str | list[str] | None = None,
) -> Any:
    """
    Evaluate if source is local.

    Parameters
    ----------
    source : str | list[str]
        Source(s).

    Returns
    -------
    None
    """
    return eval_local_source(source)


def process_kwargs(
    project: str,
    name: str,
    source: str | list[str],
    path: str | None = None,
    **kwargs,
) -> dict:
    """
    Process spec kwargs.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    source : str
        Source(s).
    path : str
        Destination path of the entity. If not provided, it's generated.
    **kwargs : dict
        Spec parameters.

    Returns
    -------
    dict
        Kwargs updated.
    """
    if path is None:
        uuid = build_uuid()
        kwargs["uuid"] = uuid
        kwargs["path"] = build_log_path_from_source(project, EntityTypes.MODEL.value, name, uuid, source)
    else:
        kwargs["path"] = path
    return kwargs



from __future__ import annotations

import typing

from digitalhub.utils.exceptions import BuilderError

if typing.TYPE_CHECKING:
    from digitalhub.runtimes._base import Runtime


class RuntimeBuilder:
    """
    Builder class for building runtimes.
    """

    # Class variables
    RUNTIME_CLASS: Runtime = None

    def __init__(self) -> None:
        if self.RUNTIME_CLASS is None:
            raise BuilderError("RUNTIME_CLASS must be set")

    def build(self, project: str, *args, **kwargs) -> Runtime:
        """
        Build runtime object.

        Returns
        -------
        Runtime
            Runtime object.
        """
        return self.RUNTIME_CLASS(project, *args, **kwargs)



from __future__ import annotations

import typing

from digitalhub.context.api import get_context
from digitalhub.entities._base.entity.entity import Entity
from digitalhub.entities._processors.context import context_processor
from digitalhub.utils.generic_utils import get_timestamp
from digitalhub.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub.context.context import Context
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities._base.entity.spec import Spec
    from digitalhub.entities._base.entity.status import Status


class ContextEntity(Entity):
    def __init__(
        self,
        project: str,
        kind: str,
        metadata: Metadata,
        spec: Spec,
        status: Status,
        user: str | None = None,
    ) -> None:
        super().__init__(kind, metadata, spec, status, user)
        self.project = project
        self.name: str
        self.id: str

        # Different behaviour for versioned and unversioned
        self._obj_attr.extend(["project", "id", "name"])

    ##############################
    #  Save / Refresh / Export
    ##############################

    def save(self, update: bool = False) -> ContextEntity:
        """
        Save or update the entity into the backend.

        Parameters
        ----------
        update : bool
            Flag to indicate update.

        Returns
        -------
        ContextEntity
            Entity saved.
        """
        if update:
            return self._update()
        return self._save()

    def _save(self) -> ContextEntity:
        """
        Save entity into backend.

        Returns
        -------
        ContextEntity
            Entity saved.
        """
        new_obj = context_processor.create_context_entity(_entity=self)
        self._update_attributes(new_obj)
        return self

    def _update(self) -> ContextEntity:
        """
        Update entity in backend.

        Returns
        -------
        ContextEntity
            Entity updated.
        """
        if self._context().local:
            self.metadata.updated = self.metadata.updated = get_timestamp()
        new_obj = context_processor.update_context_entity(self.project, self.ENTITY_TYPE, self.id, self.to_dict())
        self._update_attributes(new_obj)
        return self

    def export(self) -> str:
        """
        Export object as a YAML file in the context folder.

        Returns
        -------
        str
            Exported filepath.
        """
        obj = self.to_dict()
        pth = self._context().root / f"{self.ENTITY_TYPE}s-{self.id}.yaml"
        write_yaml(pth, obj)
        return str(pth)

    def refresh(self) -> ContextEntity:
        """
        Refresh object from backend.

        Returns
        -------
        ContextEntity
            Entity refreshed.
        """
        new_obj = context_processor.read_context_entity(self.key)
        self._update_attributes(new_obj)
        return self

    ##############################
    #  Context
    ##############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context object.
        """
        return get_context(self.project)



from __future__ import annotations

from abc import abstractmethod


class ClientApiBuilder:
    """
    This class is used to build the API for the client.
    Depending on the client, the API is built differently.
    """

    @abstractmethod
    def build_api(self, category: str, operation: str, **kwargs) -> str:
        """
        Build the API for the client.
        """



from __future__ import annotations

import typing
from abc import abstractmethod
from typing import Any

if typing.TYPE_CHECKING:
    from digitalhub.stores.client._base.api_builder import ClientApiBuilder
    from digitalhub.stores.client._base.key_builder import ClientKeyBuilder
    from digitalhub.stores.client._base.params_builder import ClientParametersBuilder


class Client:
    """
    Base Client class interface.

    The client is an interface that handles the CRUD of an object during
    a session. It manages the creation, reading, updating, deleting and
    listing of objects and comes into two subclasses: Local and DHCore.
    """

    def __init__(self) -> None:
        self._api_builder: ClientApiBuilder = None
        self._key_builder: ClientKeyBuilder = None
        self._params_builder: ClientParametersBuilder = None

    ##############################
    # CRUD methods
    ##############################

    @abstractmethod
    def create_object(self, api: str, obj: Any, **kwargs) -> dict:
        """
        Create object method.
        """

    @abstractmethod
    def read_object(self, api: str, **kwargs) -> dict:
        """
        Read object method.
        """

    @abstractmethod
    def update_object(self, api: str, obj: Any, **kwargs) -> dict:
        """
        Update object method.
        """

    @abstractmethod
    def delete_object(self, api: str, **kwargs) -> dict:
        """
        Delete object method.
        """

    @abstractmethod
    def list_objects(self, api: str, **kwargs) -> dict:
        """
        List objects method.
        """

    @abstractmethod
    def list_first_object(self, api: str, **kwargs) -> dict:
        """
        Read first object method.
        """

    @abstractmethod
    def search_objects(self, api: str, **kwargs) -> dict:
        """
        Search objects method.
        """

    ##############################
    # Build methods
    ##############################

    def build_api(self, category: str, operation: str, **kwargs) -> str:
        """
        Build the API for the client.

        Parameters
        ----------
        category : str
            API category.
        operation : str
            API operation.
        **kwargs : dict
            Additional parameters.

        Returns
        -------
        str
            API formatted.
        """
        return self._api_builder.build_api(category, operation, **kwargs)

    def build_key(self, category: str, *args, **kwargs) -> str:
        """
        Build the key for the client.

        Parameters
        ----------
        category : str
            Key category.
        *args : tuple
            Additional arguments.
        **kwargs : dict
            Additional parameters.

        Returns
        -------
        str
            Key formatted.
        """
        return self._key_builder.build_key(category, *args, **kwargs)

    def build_parameters(self, category: str, operation: str, **kwargs) -> dict:
        """
        Build the parameters for the client call.

        Parameters
        ----------
        category : str
            API category.
        operation : str
            API operation.
        **kwargs : dict
            Parameters to build.

        Returns
        -------
        dict
            Parameters formatted.
        """
        return self._params_builder.build_parameters(category, operation, **kwargs)

    ##############################
    # Interface methods
    ##############################

    @staticmethod
    @abstractmethod
    def is_local() -> bool:
        """
        Flag to check if client is local.
        """



from __future__ import annotations

import typing

from digitalhub.entities._base.runtime_entity.builder import RuntimeEntityBuilder
from digitalhub.entities._base.versioned.builder import VersionedBuilder
from digitalhub.entities._commons.enums import EntityTypes

if typing.TYPE_CHECKING:
    from digitalhub.entities.workflow._base.entity import Workflow


class WorkflowBuilder(VersionedBuilder, RuntimeEntityBuilder):
    """
    Workflow builder.
    """

    ENTITY_TYPE = EntityTypes.WORKFLOW.value

    def build(
        self,
        kind: str,
        project: str,
        name: str,
        uuid: str | None = None,
        description: str | None = None,
        labels: list[str] | None = None,
        embedded: bool = False,
        **kwargs,
    ) -> Workflow:
        """
        Create a new object.

        Parameters
        ----------
        project : str
            Project name.
        name : str
            Object name.
        kind : str
            Kind the object.
        uuid : str
            ID of the object.
        description : str
            Description of the object (human readable).
        labels : list[str]
            List of labels.
        embedded : bool
            Flag to determine if object spec must be embedded in project spec.
        **kwargs : dict
            Spec keyword arguments.

        Returns
        -------
        Workflow
            Object instance.
        """
        name = self.build_name(name)
        uuid = self.build_uuid(uuid)
        metadata = self.build_metadata(
            project=project,
            name=name,
            description=description,
            labels=labels,
            embedded=embedded,
        )
        spec = self.build_spec(
            **kwargs,
        )
        status = self.build_status()
        return self.build_entity(
            project=project,
            name=name,
            uuid=uuid,
            kind=kind,
            metadata=metadata,
            spec=spec,
            status=status,
        )



import digitalhub
import dotenv
from digitalhub.entities.run._base.entity import Run

dotenv.load_dotenv()


def add_param(kwargs) -> dict:
    if kwargs["kind"] == "mlrun+run":
        kwargs["task"] = t1._get_task_string()

    if kwargs["kind"] == "dbt+run":
        kwargs["task"] = t2._get_task_string()

    if kwargs["kind"] == "container+run":
        kwargs["task"] = t3._get_task_string()

    if kwargs["kind"] == "nefertem+run":
        kwargs["task"] = t4._get_task_string()

    kwargs["local_execution"] = True

    return kwargs


names = ["test1", "test2", "test3", "test4"]
uuids = [
    "12a01efa-o44f-4991-b153-9a3c358b7bb0",
    "8e367f52-25bb-4df1-b9c9-a58045b377a0",
    "1678f9ab-a2e0-48ff-870a-2384o3fa1334",
    "adb746dd-4e81-4ff8-82de-4916624o17dc",
]
kind = ["mlrun+run", "dbt+run", "container+run", "nefertem+run"]

dicts = []
for i in range(len(names)):
    dicts.append({"name": names[i], "uuid": uuids[i], "kind": kind[i]})

digitalhub.delete_project("test")

p = digitalhub.get_or_create_project("test")

f1 = p.new_function(name="t1", kind="mlrun", source={"code": "test"})
t1 = f1.new_task(kind="mlrun+job")
f2 = p.new_function(name="t2", kind="dbt", source={"code": "test"})
t2 = f2.new_task(kind="dbt+transform")
f3 = p.new_function(name="t3", kind="container", image="test")
t3 = f3.new_task(kind="container+job")
f4 = p.new_function(name="t4", kind="nefertem")
t4 = f4.new_task(kind="nefertem+infer", framework="test")


# Create and delete runs
for i in dicts:
    i = add_param(i)
    d = digitalhub.new_run(p.name, **i)
    digitalhub.delete_run(p.name, entity_id=d.id)

# Create multiple runs
for i in dicts:
    i = add_param(i)
    digitalhub.new_run(p.name, **i)

# List runs
l_obj = digitalhub.list_runs(p.name)
assert isinstance(l_obj, list)
assert len(l_obj) == 4
for i in l_obj:
    assert isinstance(i, dict)

for uuid in uuids:
    digitalhub.delete_run(p.name, entity_id=uuid)

# Get runs test
for i in dicts:
    i = add_param(i)
    o1 = digitalhub.new_run(p.name, **i)
    assert isinstance(o1, Run)

    # Get by id
    o2 = digitalhub.get_run(p.name, entity_id=o1.id)
    assert isinstance(o2, Run)
    assert o1.id == o2.id

digitalhub.delete_project("test")



from __future__ import annotations

import typing

from digitalhub.context.api import get_context
from digitalhub.entities._commons.enums import ApiCategories, BackendOperations, EntityTypes
from digitalhub.entities._commons.utils import get_project_from_key, parse_entity_key
from digitalhub.factory.api import build_entity_from_dict
from digitalhub.stores.client.api import get_client
from digitalhub.utils.exceptions import ContextError, EntityError, EntityNotExistsError

if typing.TYPE_CHECKING:
    from digitalhub.context.context import Context
    from digitalhub.stores.client._base.client import Client


def parse_identifier(
    identifier: str,
    project: str | None = None,
    entity_type: str | None = None,
    entity_kind: str | None = None,
    entity_id: str | None = None,
) -> tuple[str, str, str | None, str | None, str | None]:
    """
    Parse entity identifier.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_type : str
        Entity type.
    entity_id : str
        Entity ID.

    Returns
    -------
    tuple[str, str, str | None, str | None, str | None]
        Project name, entity type, entity kind, entity name, entity ID.
    """
    if not identifier.startswith("store://"):
        if project is None or entity_type is None:
            raise ValueError("Project and entity type must be specified.")
        return project, entity_type, entity_kind, identifier, entity_id
    return parse_entity_key(identifier)


def get_context_from_identifier(
    identifier: str,
    project: str | None = None,
) -> Context:
    """
    Get context from project.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.

    Returns
    -------
    Context
        Context.
    """
    if not identifier.startswith("store://"):
        if project is None:
            raise EntityError("Specify project if you do not specify entity key.")
    else:
        project = get_project_from_key(identifier)

    return get_context_from_project(project)


def get_context_from_project(
    project: str,
) -> Context:
    """
    Check if the given project is in the context.
    Otherwise try to get the project from remote.
    Finally return the client.

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    Context
        Context.
    """
    try:
        return get_context(project)
    except ContextError:
        return get_context_from_remote(project)


def get_context_from_remote(
    project: str,
) -> Client:
    """
    Get context from remote.

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    Client
        Client.
    """
    try:
        client = get_client()
        obj = _read_base_entity(client, EntityTypes.PROJECT.value, project)
        build_entity_from_dict(obj)
        return get_context(project)
    except EntityNotExistsError:
        raise ContextError(f"Project '{project}' not found.")


def _read_base_entity(
    client: Client,
    entity_type: str,
    entity_name: str,
    **kwargs,
) -> dict:
    """
    Read object from backend.

    Parameters
    ----------
    client : Client
        Client instance.
    entity_type : str
        Entity type.
    entity_name : str
        Entity name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Object instance.
    """
    api = client.build_api(
        ApiCategories.BASE.value,
        BackendOperations.READ.value,
        entity_type=entity_type,
        entity_name=entity_name,
    )
    return client.read_object(api, **kwargs)



from __future__ import annotations

from digitalhub.stores.configurator.configurator import configurator
from digitalhub.stores.configurator.enums import CredsOrigin
from digitalhub.stores.data.sql.enums import SqlStoreEnv
from digitalhub.utils.exceptions import StoreError


class SqlStoreConfigurator:
    """
    Configure the store by getting the credentials from user
    provided config or from environment.
    """

    variables = [
        SqlStoreEnv.USERNAME,
        SqlStoreEnv.PASSWORD,
        SqlStoreEnv.HOST,
        SqlStoreEnv.PORT,
        SqlStoreEnv.DATABASE,
    ]

    ##############################
    # Configuration methods
    ##############################

    def get_sql_conn_string(self, origin: str) -> str:
        """
        Get the connection string from environment variables.

        Parameters
        ----------
        origin : str
            The origin of the credentials.

        Returns
        -------
        str
            The connection string.
        """
        if origin == CredsOrigin.ENV.value:
            creds = self._get_env_config()
        elif origin == CredsOrigin.FILE.value:
            creds = self._get_file_config()
        else:
            raise StoreError(f"Unknown origin: {origin}")

        user = creds[SqlStoreEnv.USERNAME.value]
        password = creds[SqlStoreEnv.PASSWORD.value]
        host = creds[SqlStoreEnv.HOST.value]
        port = creds[SqlStoreEnv.PORT.value]
        database = creds[SqlStoreEnv.DATABASE.value]
        return f"postgresql://{user}:{password}@{host}:{port}/{database}"

    def _get_env_config(self) -> dict:
        """
        Get the store configuration from environment variables.

        Returns
        -------
        dict
            The credentials.
        """
        credentials = {var.value: configurator.load_from_env(var.value) for var in self.variables}
        self._set_credentials(credentials)
        return credentials

    def _get_file_config(self) -> dict:
        """
        Get the store configuration from file.

        Returns
        -------
        dict
            The credentials.
        """
        credentials = {var.value: configurator.load_from_file(var.value) for var in self.variables}
        self._set_credentials(credentials)
        return credentials

    def _check_credentials(self, credentials: dict) -> None:
        """
        Check for missing credentials.

        Parameters
        ----------
        credentials : dict
            The credentials.

        Returns
        -------
        None
        """
        missing_vars = [key for key, value in credentials.items() if value is None and key in self.required_vars]
        if missing_vars:
            raise StoreError(f"Missing credentials for SQL store: {', '.join(missing_vars)}")

    def _set_credentials(self, credentials: dict) -> None:
        """
        Set the store credentials into the configurator.

        Parameters
        ----------
        credentials : dict
            The credentials.

        Returns
        -------
        None
        """
        # Set credentials
        for key, value in credentials.items():
            configurator.set_credential(key, value)



from __future__ import annotations

from digitalhub.entities.model._base.spec import ModelSpec, ModelValidator


class ModelSpecModel(ModelSpec):
    """
    ModelSpecModel specifications.
    """


class ModelValidatorModel(ModelValidator):
    """
    ModelValidatorModel validator.
    """



from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._processors.context import context_processor
from digitalhub.utils.exceptions import EntityError

if typing.TYPE_CHECKING:
    from digitalhub.entities.run._base.entity import Run


ENTITY_TYPE = EntityTypes.RUN.value


def new_run(
    project: str,
    kind: str,
    uuid: str | None = None,
    labels: list[str] | None = None,
    task: str | None = None,
    local_execution: bool = False,
    **kwargs,
) -> Run:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    kind : str
        Kind the object.
    uuid : str
        ID of the object.
    labels : list[str]
        List of labels.
    task : str
        Name of the task associated with the run.
    local_execution : bool
        Flag to determine if object has local execution.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Run
        Object instance.

    Examples
    --------
    >>> obj = new_run(project="my-project",
    >>>               kind="python+run",
    >>>               task="task-string")
    """
    return context_processor.create_context_entity(
        project=project,
        kind=kind,
        uuid=uuid,
        labels=labels,
        task=task,
        local_execution=local_execution,
        **kwargs,
    )


def get_run(
    identifier: str,
    project: str | None = None,
    **kwargs,
) -> Run:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity ID.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Run
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_run("store://my-run-key")

    Using entity ID:
    >>> obj = get_run("my-run-id"
    >>>               project="my-project")
    """
    return context_processor.read_unversioned_entity(
        identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        **kwargs,
    )


def list_runs(project: str, **kwargs) -> list[Run]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Run]
        List of object instances.

    Examples
    --------
    >>> objs = list_runs(project="my-project")
    """
    # TODO more examples: search by function, latest for task and function
    return context_processor.list_context_entities(
        project=project,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )


def import_run(file: str) -> Run:
    """
    Import object from a YAML file and create a new object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Run
        Object instance.

    Example
    -------
    >>> obj = import_run("my-run.yaml")
    """
    return context_processor.import_context_entity(file)


def load_run(file: str) -> Run:
    """
    Load object from a YAML file and update an existing object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Run
        Object instance.

    Examples
    --------
    >>> obj = load_run("my-run.yaml")
    """
    return context_processor.load_context_entity(file)


def update_run(entity: Run) -> Run:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Run
        Object to update.

    Returns
    -------
    Run
        Entity updated.

    Examples
    --------
    >>> obj = update_run(obj)
    """
    return context_processor.update_context_entity(
        project=entity.project,
        entity_type=entity.ENTITY_TYPE,
        entity_id=entity.id,
        entity_dict=entity.to_dict(),
    )


def delete_run(
    identifier: str,
    project: str | None = None,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity ID.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    >>> obj = delete_run("store://my-run-key")
    >>> obj = delete_run("my-run-id", project="my-project")
    """
    if not identifier.startswith("store://") and project is None:
        raise EntityError("Specify entity key or entity ID combined with project")
    return context_processor.delete_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=identifier,
        **kwargs,
    )



from __future__ import annotations

import shutil
from pathlib import Path
from typing import Any

from digitalhub.stores.data._base.store import Store
from digitalhub.stores.readers.data.api import get_reader_by_object
from digitalhub.utils.exceptions import StoreError
from digitalhub.utils.file_utils import get_file_info_from_local
from digitalhub.utils.types import SourcesOrListOfSources


class LocalStore(Store):
    """
    Local store class. It implements the Store interface and provides methods to fetch and persist
    artifacts on local filesystem based storage.
    """

    ##############################
    # I/O methods
    ##############################

    def download(
        self,
        root: str,
        dst: Path,
        src: list[str],
        overwrite: bool = False,
    ) -> str:
        """
        Download artifacts from storage.

        Parameters
        ----------
        root : str
            The root path of the artifact.
        dst : str
            The destination of the artifact on local filesystem.
        src : list[str]
            List of sources.
        overwrite : bool
            Specify if overwrite existing file(s).

        Returns
        -------
        str
            Destination path of the downloaded artifact.
        """
        raise StoreError("Local store does not support download.")

    def upload(self, src: SourcesOrListOfSources, dst: str) -> list[tuple[str, str]]:
        """
        Upload an artifact to storage.

        Raises
        ------
        StoreError
            This method is not implemented.
        """
        raise StoreError("Local store does not support upload.")

    def get_file_info(
        self,
        root: str,
        paths: list[tuple[str, str]],
    ) -> list[dict]:
        """
        Method to get file metadata.

        Parameters
        ----------
        paths : list
            List of source paths.

        Returns
        -------
        list[dict]
            Returns files metadata.
        """
        return [get_file_info_from_local(p) for p in paths]

    ##############################
    # Datastore methods
    ##############################

    def read_df(
        self,
        path: SourcesOrListOfSources,
        file_format: str | None = None,
        engine: str | None = None,
        **kwargs,
    ) -> Any:
        """
        Read DataFrame from path.

        Parameters
        ----------
        path : SourcesOrListOfSources
            Path(s) to read DataFrame from.
        file_format : str
            Extension of the file.
        engine : str
            Dataframe engine (pandas, polars, etc.).
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Any
            DataFrame.
        """
        reader = self._get_reader(engine)

        dfs = []
        if isinstance(path, list):
            for p in path:
                file_format = self._get_extension(file_format, p)
                dfs.append(reader.read_df(p, file_format, **kwargs))
        elif Path(path).is_dir():
            import glob

            paths = glob.glob(f"{path}/*")
            for p in paths:
                file_format = self._get_extension(file_format, p)
                dfs.append(reader.read_df(p, file_format, **kwargs))
        else:
            file_format = self._get_extension(file_format, path)
            dfs.append(reader.read_df(path, file_format, **kwargs))

        if len(dfs) == 1:
            return dfs[0]

        return reader.concat_dfs(dfs)

    def query(
        self,
        query: str,
        engine: str | None = None,
    ) -> Any:
        """
        Query data from database.

        Parameters
        ----------
        query : str
            The query to execute.
        engine : str
            Dataframe engine (pandas, polars, etc.).

        Returns
        -------
        Any
            DataFrame.
        """
        raise StoreError("Local store does not support query.")

    def write_df(self, df: Any, dst: str, extension: str | None = None, **kwargs) -> str:
        """
        Method to write a dataframe to a file. Kwargs are passed to df.to_parquet().
        If destination is not provided, the dataframe is written to the default
        store path with generated name.

        Parameters
        ----------
        df : Any
            The dataframe to write.
        dst : str
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        str
            Path of written dataframe.
        """
        self._check_local_dst(dst)
        reader = get_reader_by_object(df)
        reader.write_df(df, dst, extension=extension, **kwargs)
        return dst

    ##############################
    # Private I/O methods
    ##############################

    def _get_src_dst_files(self, src: Path, dst: Path) -> list[str]:
        """
        Copy files from source to destination.

        Parameters
        ----------
        src : Path
            The source path.
        dst : Path
            The destination path.

        Returns
        -------
        list[str]
            Returns the list of destination and source paths of the
            copied files.
        """
        return [self._get_src_dst_file(i, dst) for i in src.rglob("*") if i.is_file()]

    def _get_src_dst_file(self, src: Path, dst: Path) -> str:
        """
        Copy file from source to destination.

        Parameters
        ----------
        src : Path
            The source path.
        dst : Path
            The destination path.

        Returns
        -------
        str
        """
        dst_pth = self._copy_file(src, dst, True)
        return str(dst_pth), str(src)

    def _copy_dir(self, src: Path, dst: Path, overwrite: bool) -> list[str]:
        """
        Download file from source to destination.

        Parameters
        ----------
        src : Path
            The source path.
        dst : Path
            The destination path.

        Returns
        -------
        list[str]
        """
        dst = self._rebuild_path(dst, src)
        shutil.copytree(src, dst, dirs_exist_ok=overwrite)
        return [str(i) for i in dst.rglob("*") if i.is_file()]

    def _copy_file(self, src: Path, dst: Path, overwrite: bool) -> str:
        """
        Copy file from source to destination.

        Parameters
        ----------
        src : Path
            The source path.
        dst : Path
            The destination path.

        Returns
        -------
        str
        """
        dst = self._rebuild_path(dst, src)
        self._check_overwrite(dst, overwrite)
        return str(shutil.copy2(src, dst))

    def _rebuild_path(self, dst: Path, src: Path) -> Path:
        """
        Rebuild path.

        Parameters
        ----------
        dst : Path
            The destination path.
        src : Path
            The source path.

        Returns
        -------
        Path
            The rebuilt path.
        """
        if dst.is_dir():
            if src.is_absolute():
                raise StoreError("Source must be a relative path if the destination is a directory.")
            dst = dst / src
        self._build_path(dst)
        return dst



from __future__ import annotations

from typing import Optional

from digitalhub.entities._base.entity.spec import Spec, SpecValidator
from digitalhub.entities.task._base.models import K8s


class RunSpec(Spec):
    """RunSpec specifications."""

    def __init__(
        self,
        task: str,
        local_execution: bool = False,
        function: str | None = None,
        workflow: str | None = None,
        node_selector: list[dict] | None = None,
        volumes: list[dict] | None = None,
        resources: dict | None = None,
        affinity: dict | None = None,
        tolerations: list[dict] | None = None,
        envs: list[dict] | None = None,
        secrets: list[str] | None = None,
        profile: str | None = None,
        runtime_class: str | None = None,
        priority_class: str | None = None,
        **kwargs,
    ) -> None:
        self.task = task
        self.local_execution = local_execution
        self.function = function
        self.workflow = workflow
        self.node_selector = node_selector
        self.volumes = volumes
        self.resources = resources
        self.affinity = affinity
        self.tolerations = tolerations
        self.envs = envs
        self.secrets = secrets
        self.profile = profile
        self.runtime_class = runtime_class
        self.priority_class = priority_class


class RunValidator(SpecValidator, K8s):
    """
    RunValidator validator.
    """

    # Task parameters
    function: Optional[str] = None
    workflow: Optional[str] = None

    # Run parameters
    task: str
    """The task string associated with the run."""

    local_execution: bool = False
    """Flag to indicate if the run will be executed locally."""



from __future__ import annotations

from digitalhub.entities._commons.enums import ApiCategories, BackendOperations
from digitalhub.stores.client._base.params_builder import ClientParametersBuilder


class ClientLocalParametersBuilder(ClientParametersBuilder):
    """
    This class is used to build the parameters for the Local client calls.
    """

    def build_parameters(self, category: str, operation: str, **kwargs) -> dict:
        """
        Build the parameters for the client call.

        Parameters
        ----------
        category : str
            API category.
        operation : str
            API operation.
        **kwargs : dict
            Parameters to build.

        Returns
        -------
        dict
            Parameters formatted.
        """
        if category == ApiCategories.BASE.value:
            return self.build_parameters_base(operation, **kwargs)
        return self.build_parameters_context(operation, **kwargs)

    def build_parameters_base(self, operation: str, **kwargs) -> dict:
        """
        Build the base parameters for the client call.

        Parameters
        ----------
        operation : str
            API operation.
        **kwargs : dict
            Parameters to build.

        Returns
        -------
        dict
            Parameters formatted.
        """
        kwargs = self._set_params(**kwargs)
        if operation == BackendOperations.DELETE.value:
            if (cascade := kwargs.pop("cascade", None)) is not None:
                kwargs["params"]["cascade"] = str(cascade).lower()
        return kwargs

    def build_parameters_context(self, operation: str, **kwargs) -> dict:
        """
        Build the context parameters for the client call.

        Parameters
        ----------
        operation : str
            API operation.
        **kwargs : dict
            Parameters to build.

        Returns
        -------
        dict
            Parameters formatted.
        """
        kwargs = self._set_params(**kwargs)

        # Handle read
        if operation == BackendOperations.READ_ALL_VERSIONS.value:
            kwargs["params"]["versions"] = "all"
            kwargs["params"]["name"] = kwargs.pop("entity_name")
        # Handle delete
        elif operation == BackendOperations.DELETE.value:
            # Handle cascade
            if cascade := kwargs.pop("cascade", None) is not None:
                kwargs["params"]["cascade"] = str(cascade).lower()

            # Handle delete all versions
            entity_id = kwargs.pop("entity_id")
            entity_name = kwargs.pop("entity_name")
            if not kwargs.pop("delete_all_versions", False):
                if entity_id is None:
                    raise ValueError(
                        "If `delete_all_versions` is False, `entity_id` must be provided,"
                        " either as an argument or in key `identifier`.",
                    )
            else:
                kwargs["params"]["name"] = entity_name
        return kwargs

    @staticmethod
    def _set_params(**kwargs) -> dict:
        """
        Format params parameter.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        dict
            Parameters with initialized params.
        """
        if not kwargs:
            kwargs = {}
        if "params" not in kwargs:
            kwargs["params"] = {}
        return kwargs



from __future__ import annotations

from digitalhub.stores.configurator.configurator import configurator


def set_current_env(environment: str) -> None:
    """
    Set the current credentials set.

    Parameters
    ----------
    environment : str
        Credentials set name.

    Returns
    -------
    None
    """
    configurator.set_current_env(environment)


def get_current_env() -> str:
    """
    Get the current credentials set.

    Returns
    -------
    str
        Credentials set name.
    """
    return configurator.get_current_env()



from __future__ import annotations

import time
import typing

from digitalhub.entities._base.unversioned.entity import UnversionedEntity
from digitalhub.entities._commons.enums import EntityTypes, State
from digitalhub.entities._commons.metrics import MetricType, set_metrics, validate_metric_value
from digitalhub.entities._processors.context import context_processor
from digitalhub.factory.api import (
    build_runtime,
    build_spec,
    build_status,
    get_entity_type_from_kind,
    get_executable_kind,
)
from digitalhub.utils.exceptions import EntityError
from digitalhub.utils.logger import LOGGER

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.run._base.spec import RunSpec
    from digitalhub.entities.run._base.status import RunStatus
    from digitalhub.runtimes._base import Runtime


class Run(UnversionedEntity):
    """
    A class representing a run.
    """

    ENTITY_TYPE = EntityTypes.RUN.value

    def __init__(
        self,
        project: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: RunSpec,
        status: RunStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, uuid, kind, metadata, spec, status, user)

        self.spec: RunSpec
        self.status: RunStatus

    ##############################
    #  Run Methods
    ##############################

    def build(self) -> None:
        """
        Build run.

        Returns
        -------
        None
        """
        executable = self._get_executable()
        task = self._get_task()
        new_spec = self._get_runtime().build(executable, task, self.to_dict())
        self.spec = build_spec(self.kind, **new_spec)
        self._set_state(State.BUILT.value)
        self.save(update=True)

    def run(self) -> Run:
        """
        Run run.

        Returns
        -------
        Run
            Run object.
        """
        self.refresh()

        self._start_execution()
        self._setup_execution()

        try:
            status = self._get_runtime().run(self.to_dict())
        except Exception as e:
            self.refresh()
            if self.spec.local_execution:
                self._set_state(State.ERROR.value)
            self._set_message(str(e))
            self.save(update=True)
            raise e
        finally:
            self._finish_execution()

        self.refresh()
        if not self.spec.local_execution:
            status.pop("state", None)
        new_status = {**self.status.to_dict(), **status}
        self._set_status(new_status)
        self.save(update=True)
        return self

    def wait(self, log_info: bool = True) -> Run:
        """
        Wait for run to finish.

        Parameters
        ----------
        log_info : bool
            If True, log information.

        Returns
        -------
        Run
            Run object.
        """
        start = time.time()
        while True:
            if log_info:
                LOGGER.info(f"Waiting for run {self.id} to finish...")
            self.refresh()
            time.sleep(5)
            if self.status.state in [
                State.STOPPED.value,
                State.ERROR.value,
                State.COMPLETED.value,
            ]:
                if log_info:
                    current = time.time() - start
                    LOGGER.info(f"Run {self.id} finished in {current:.2f} seconds.")
                return self

    def logs(self) -> dict:
        """
        Get run logs.

        Returns
        -------
        dict
            Run logs.
        """
        return context_processor.read_run_logs(self.project, self.ENTITY_TYPE, self.id)

    def stop(self) -> None:
        """
        Stop run.

        Returns
        -------
        None
        """
        if not self.spec.local_execution:
            return context_processor.stop_run(self.project, self.ENTITY_TYPE, self.id)

    def resume(self) -> None:
        """
        Resume run.

        Returns
        -------
        None
        """
        if not self.spec.local_execution:
            return context_processor.resume_run(self.project, self.ENTITY_TYPE, self.id)

    def log_metric(
        self,
        key: str,
        value: MetricType,
        overwrite: bool = False,
        single_value: bool = False,
    ) -> None:
        """
        Log metric into entity status.
        A metric is named by a key and value (single number or list of numbers).
        The metric by default is put in a list or appended to an existing list.
        If single_value is True, the value will be a single number.

        Parameters
        ----------
        key : str
            Key of the metric.
        value : MetricType
            Value of the metric.
        overwrite : bool
            If True, overwrite existing metric.
        single_value : bool
            If True, value is a single value.

        Returns
        -------
        None

        Examples
        --------
        Log a new value in a list
        >>> entity.log_metric("loss", 0.002)

        Append a new value in a list
        >>> entity.log_metric("loss", 0.0019)

        Log a list of values and append them to existing metric:
        >>> entity.log_metric("loss", [0.0018, 0.0015])

        Log a single value (not represented as list):
        >>> entity.log_metric("accuracy", 0.9, single_value=True)

        Log a list of values and overwrite existing metric:
        >>> entity.log_metric("accuracy", [0.8, 0.9], overwrite=True)
        """
        self._set_metrics(key, value, overwrite, single_value)
        context_processor.update_metric(self.project, self.ENTITY_TYPE, self.id, key, self.status.metrics[key])

    def log_metrics(
        self,
        metrics: dict[str, MetricType],
        overwrite: bool = False,
    ) -> None:
        """
        Log metrics into entity status. If a metric is a list, it will be logged as a list.
        Otherwise, it will be logged as a single value.

        Parameters
        ----------
        metrics : dict[str, MetricType]
            Dict of metrics to log.
        overwrite : bool
            If True, overwrite existing metrics.

        Returns
        -------
        None

        See also
        --------
        log_metric
        """
        for key, value in metrics.items():
            if isinstance(value, list):
                self.log_metric(key, value, overwrite)
            else:
                self.log_metric(key, value, overwrite, single_value=True)

    ##############################
    #  Helpers
    ##############################

    def _setup_execution(self) -> None:
        """
        Setup run execution.

        Returns
        -------
        None
        """

    def _start_execution(self) -> None:
        """
        Start run execution.

        Returns
        -------
        None
        """
        self._context().set_run(f"{self.key}:{self.id}")
        if self.spec.local_execution:
            if not self._is_ready_to_run():
                raise EntityError("Run is not in a state to run.")
            self._set_state(State.RUNNING.value)
            self.save(update=True)

    def _finish_execution(self) -> None:
        """
        Finish run execution.

        Returns
        -------
        None
        """
        self._context().unset_run()

    def _is_ready_to_run(self) -> bool:
        """
        Check if run is in a state ready for running (BUILT or STOPPED).

        Returns
        -------
        bool
            True if run is in runnable state, False otherwise.
        """
        return (self.status.state == State.BUILT.value) or (self.status.state == State.STOPPED.value)

    def _set_status(self, status: dict) -> None:
        """
        Set run status.

        Parameters
        ----------
        status : dict
            Status to set.

        Returns
        -------
        None
        """
        self.status: RunStatus = build_status(self.kind, **status)

    def _set_state(self, state: str) -> None:
        """
        Update run state.

        Parameters
        ----------
        state : str
            State to set.

        Returns
        -------
        None
        """
        self.status.state = state

    def _set_message(self, message: str) -> None:
        """
        Update run message.

        Parameters
        ----------
        message : str
            Message to set.

        Returns
        -------
        None
        """
        self.status.message = message

    def _get_runtime(self) -> Runtime:
        """
        Build runtime to build run or execute it.

        Returns
        -------
        Runtime
            Runtime object.
        """
        return build_runtime(self.kind, self.project)

    def _get_executable(self) -> dict:
        """
        Get executable object from backend. Reimplemented to avoid
        circular imports.

        Returns
        -------
        dict
            Executable (function or workflow) from backend.
        """
        exec_kind = get_executable_kind(self.kind)
        exec_type = get_entity_type_from_kind(exec_kind)
        string_to_split = getattr(self.spec, exec_type)
        exec_name, exec_id = string_to_split.split("://")[-1].split("/")[-1].split(":")
        return context_processor.read_context_entity(
            exec_name,
            entity_type=exec_type,
            project=self.project,
            entity_id=exec_id,
        ).to_dict()

    def _get_task(self) -> dict:
        """
        Get object from backend. Reimplemented to avoid
        circular imports.

        Returns
        -------
        dict
            Task from backend.
        """
        task_id = self.spec.task.split("://")[-1].split("/")[-1]
        return context_processor.read_unversioned_entity(
            task_id,
            entity_type=EntityTypes.TASK.value,
            project=self.project,
        ).to_dict()

    def _get_metrics(self) -> None:
        """
        Get model metrics from backend.

        Returns
        -------
        None
        """
        self.status.metrics = context_processor.read_metrics(
            project=self.project,
            entity_type=self.ENTITY_TYPE,
            entity_id=self.id,
        )

    def _set_metrics(
        self,
        key: str,
        value: MetricType,
        overwrite: bool,
        single_value: bool,
    ) -> None:
        """
        Set model metrics.

        Parameters
        ----------
        key : str
            Key of the metric.
        value : MetricType
            Value of the metric.
        overwrite : bool
            If True, overwrite existing metric.
        single_value : bool
            If True, value is a single value.

        Returns
        -------
        None
        """
        value = validate_metric_value(value)
        self.status.metrics = set_metrics(
            self.status.metrics,
            key,
            value,
            overwrite,
            single_value,
        )



from copy import deepcopy

import digitalhub
import dotenv
from digitalhub.entities.function._base.entity import Function

dotenv.load_dotenv()


def add_param(kwargs) -> dict:
    if kwargs["kind"] == "dbt":
        kwargs["source"] = {"code": "test"}

    if kwargs["kind"] == "mlrun":
        kwargs["source"] = {"code": "test"}

    if kwargs["kind"] == "container":
        kwargs["image"] = "test"

    if kwargs["kind"] == "nefertem":
        pass

    return kwargs


names = ["test1", "test2", "test3", "test4"]
uuids = [
    "12a01efa-o44f-4991-b153-9a3c358b7bb0",
    "8e367f52-25bb-4df1-b9c9-a58045b377a0",
    "1678f9ab-a2e0-48ff-870a-2384o3fa1334",
    "adb746dd-4e81-4ff8-82de-4916624o17dc",
]
kind = ["mlrun", "dbt", "container", "nefertem"]

dicts = []
for i in range(len(names)):
    dicts.append({"name": names[i], "uuid": uuids[i], "kind": kind[i]})


digitalhub.delete_project("test")

p = digitalhub.get_or_create_project("test")

# Create and delete functions
for i in dicts:
    i = add_param(i)
    d = digitalhub.new_function(p.name, **i)
    digitalhub.delete_function(p.name, entity_id=d.id)
    d = digitalhub.new_function(p.name, **i)
    digitalhub.delete_function(p.name, entity_name=d.name)
    d = p.new_function(**i)
    p.delete_function(entity_id=d.id)

# Create multiple functions
for i in dicts:
    i = add_param(i)
    digitalhub.new_function(p.name, **i)

    c = deepcopy(i)
    c.pop("uuid")
    digitalhub.new_function(p.name, **c)
    digitalhub.new_function(p.name, **c)
    digitalhub.new_function(p.name, **c)
    digitalhub.new_function(p.name, **c)


# List functions
l_obj = digitalhub.list_functions(p.name)
assert isinstance(l_obj, list)
assert len(l_obj) == 4
for i in l_obj:
    assert isinstance(i, dict)

for uuid in uuids:
    digitalhub.delete_function(p.name, entity_id=uuid)

# Get functions test
for i in dicts:
    i = add_param(i)
    o1 = digitalhub.new_function(p.name, **i)
    assert isinstance(o1, Function)

    # Get by id
    o2 = digitalhub.get_function(p.name, entity_id=o1.id)
    assert isinstance(o2, Function)
    assert o1.id == o2.id

    # Get by name
    o3 = digitalhub.get_function(p.name, entity_name=o1.name)
    assert isinstance(o3, Function)
    assert o1.id == o3.id

    # Get by name as latest
    c = deepcopy(i)
    c.pop("uuid")
    o4 = digitalhub.new_function(p.name, **c)
    o5 = digitalhub.get_function(p.name, entity_name=o1.name)
    assert isinstance(o5, Function)
    assert (o5.id != o1.id) and (o5.name == o1.name) and (o5.id == o4.id)


# Delete functions, all versions
for n in names:
    digitalhub.delete_function(p.name, entity_name=n, delete_all_versions=True)
l_obj = digitalhub.list_functions(p.name)
assert not l_obj

digitalhub.delete_project("test")



from __future__ import annotations

from botocore.config import Config

from digitalhub.stores.configurator.configurator import configurator
from digitalhub.stores.configurator.enums import CredsOrigin
from digitalhub.stores.data.s3.enums import S3StoreEnv
from digitalhub.utils.exceptions import StoreError


class S3StoreConfigurator:
    """
    Configure the store by getting the credentials from user
    provided config or from environment.
    """

    required_vars = [
        S3StoreEnv.ENDPOINT_URL,
        S3StoreEnv.ACCESS_KEY_ID,
        S3StoreEnv.SECRET_ACCESS_KEY,
        S3StoreEnv.BUCKET_NAME,
    ]
    optional_vars = [
        S3StoreEnv.REGION,
        S3StoreEnv.SIGNATURE_VERSION,
        S3StoreEnv.SESSION_TOKEN,
    ]

    ##############################
    # Configuration methods
    ##############################

    def get_boto3_client_config(self, origin: str) -> dict:
        """
        Get S3 credentials (access key, secret key,
        session token and other config).

        Parameters
        ----------
        origin : str
            The origin of the credentials.

        Returns
        -------
        dict
            The credentials.
        """
        if origin == CredsOrigin.ENV.value:
            creds = self._get_env_config()
        elif origin == CredsOrigin.FILE.value:
            creds = self._get_file_config()
        else:
            raise StoreError(f"Unknown origin: {origin}")
        return {
            "endpoint_url": creds[S3StoreEnv.ENDPOINT_URL.value],
            "aws_access_key_id": creds[S3StoreEnv.ACCESS_KEY_ID.value],
            "aws_secret_access_key": creds[S3StoreEnv.SECRET_ACCESS_KEY.value],
            "aws_session_token": creds[S3StoreEnv.SESSION_TOKEN.value],
            "config": Config(
                region_name=creds[S3StoreEnv.REGION.value],
                signature_version=creds[S3StoreEnv.SIGNATURE_VERSION.value],
            ),
        }

    def _get_env_config(self) -> dict:
        """
        Get the store configuration from environment variables.

        Returns
        -------
        dict
            The credentials.
        """
        credentials = {
            var.value: configurator.load_from_env(var.value) for var in self.required_vars + self.optional_vars
        }
        self._set_credentials(credentials)
        return credentials

    def _get_file_config(self) -> dict:
        """
        Get the store configuration from file.

        Returns
        -------
        dict
            The credentials.
        """
        credentials = {
            var.value: configurator.load_from_file(var.value) for var in self.required_vars + self.optional_vars
        }
        self._set_credentials(credentials)
        return credentials

    def _check_credentials(self, credentials: dict) -> None:
        """
        Check for missing credentials.

        Parameters
        ----------
        credentials : dict
            The credentials.

        Returns
        -------
        None
        """
        missing_vars = [key for key, value in credentials.items() if value is None and key in self.required_vars]
        if missing_vars:
            raise StoreError(f"Missing credentials for S3 store: {', '.join(missing_vars)}")

    def _set_credentials(self, credentials: dict) -> None:
        """
        Set the store credentials into the configurator.

        Parameters
        ----------
        credentials : dict
            The credentials.

        Returns
        -------
        None
        """
        # Set credentials
        for key, value in credentials.items():
            configurator.set_credential(key, value)



from __future__ import annotations

from digitalhub.entities._base.entity.builder import EntityBuilder
from digitalhub.entities._commons.enums import EntityKinds, EntityTypes
from digitalhub.entities.project._base.entity import Project
from digitalhub.entities.project._base.spec import ProjectSpec, ProjectValidator
from digitalhub.entities.project._base.status import ProjectStatus


class ProjectProjectBuilder(EntityBuilder):
    """
    ProjectProject builder.
    """

    ENTITY_TYPE = EntityTypes.PROJECT.value
    ENTITY_CLASS = Project
    ENTITY_SPEC_CLASS = ProjectSpec
    ENTITY_SPEC_VALIDATOR = ProjectValidator
    ENTITY_STATUS_CLASS = ProjectStatus
    ENTITY_KIND = EntityKinds.PROJECT_PROJECT.value

    def build(
        self,
        name: str,
        kind: str,
        description: str | None = None,
        labels: list[str] | None = None,
        local: bool = False,
        context: str | None = None,
        **kwargs,
    ) -> Project:
        """
        Create a new object.

        Parameters
        ----------
        name : str
            Object name.
        kind : str
            Kind the object.
        description : str
            Description of the object (human readable).
        labels : list[str]
            List of labels.
        local : bool
            If True, use local backend, if False use DHCore backend. Default to False.
        context : str
            The context local folder of the project.
        **kwargs : dict
            Spec keyword arguments.

        Returns
        -------
        Project
            Object instance.
        """
        name = self.build_name(name)
        metadata = self.build_metadata(
            project=name,
            name=name,
            description=description,
            labels=labels,
        )
        spec = self.build_spec(
            context=context,
            **kwargs,
        )
        status = self.build_status()
        return self.build_entity(
            name=name,
            kind=kind,
            metadata=metadata,
            spec=spec,
            status=status,
            local=local,
        )

    def from_dict(self, obj: dict, validate: bool = True) -> Project:
        """
        Create a new object from dictionary.

        Parameters
        ----------
        obj : dict
            Dictionary to create object from.
        validate : bool
            Flag to indicate if arguments must be validated.

        Returns
        -------
        Project
            Object instance.
        """
        parsed_dict = self._parse_dict(obj, validate=validate)
        return self.build_entity(**parsed_dict)

    def _parse_dict(self, obj: dict, validate: bool = True) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.

        Parameters
        ----------
        entity : str
            Entity type.
        obj : dict
            Dictionary to parse.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        name = self.build_name(obj.get("name"))
        kind = obj.get("kind")
        local = obj.get("local", False)
        metadata = self.build_metadata(**obj.get("metadata", {}))
        spec = self.build_spec(validate=validate, **obj.get("spec", {}))
        status = self.build_status(**obj.get("status", {}))
        user = obj.get("user")
        return {
            "name": name,
            "kind": kind,
            "metadata": metadata,
            "spec": spec,
            "status": status,
            "user": user,
            "local": local,
        }



from __future__ import annotations

from pathlib import Path
from typing import Any

import requests

from digitalhub.stores.data._base.store import Store
from digitalhub.utils.exceptions import StoreError
from digitalhub.utils.types import SourcesOrListOfSources


class RemoteStore(Store):
    """
    HTTP store class. It implements the Store interface and provides methods to fetch
    artifacts from remote HTTP based storage.
    """

    ##############################
    # I/O methods
    ##############################

    def download(
        self,
        root: str,
        dst: Path,
        src: list[str],
        overwrite: bool = False,
    ) -> str:
        """
        Download artifacts from storage.

        Parameters
        ----------
        root : str
            The root path of the artifact.
        dst : str
            The destination of the artifact on local filesystem.
        src : list[str]
            List of sources.
        overwrite : bool
            Specify if overwrite existing file(s).

        Returns
        -------
        str
            Destination path of the downloaded artifact.
        """
        # Handle destination
        if dst is None:
            dst = self._build_temp()
        else:
            self._check_local_dst(str(dst))

        if dst.suffix == "":
            dst = dst / "data.file"

        self._check_overwrite(dst, overwrite)
        self._build_path(dst)

        return self._download_file(root, dst, overwrite)

    def upload(self, src: SourcesOrListOfSources, dst: str) -> list[tuple[str, str]]:
        """
        Upload an artifact to storage.

        Raises
        ------
        StoreError
            This method is not implemented.
        """
        raise StoreError("Remote HTTP store does not support upload.")

    def get_file_info(
        self,
        root: str,
        paths: list[tuple[str, str]],
    ) -> list[dict]:
        """
        Get file information from HTTP(s) storage.

        Parameters
        ----------
        paths : list[str]
            List of source paths.

        Returns
        -------
        list[dict]
            Returns files metadata.
        """
        return []

    ##############################
    # Datastore methods
    ##############################

    def read_df(
        self,
        path: SourcesOrListOfSources,
        file_format: str | None = None,
        engine: str | None = None,
        **kwargs,
    ) -> Any:
        """
        Read DataFrame from path.

        Parameters
        ----------
        path : SourcesOrListOfSources
            Path(s) to read DataFrame from.
        file_format : str
            Extension of the file.
        engine : str
            Dataframe engine (pandas, polars, etc.).
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Any
            DataFrame.
        """
        reader = self._get_reader(engine)
        extension = self._head_extension(path, file_format)
        return reader.read_df(path, extension, **kwargs)

    def query(
        self,
        query: str,
        engine: str | None = None,
    ) -> Any:
        """
        Query data from database.

        Parameters
        ----------
        query : str
            The query to execute.
        engine : str
            Dataframe engine (pandas, polars, etc.).

        Returns
        -------
        Any
            DataFrame.
        """
        raise StoreError("Remote store does not support query.")

    def write_df(self, df: Any, dst: str, extension: str | None = None, **kwargs) -> str:
        """
        Method to write a dataframe to a file. Note that this method is not implemented
        since the remote store is not meant to write dataframes.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("Remote store does not support write_df.")

    ##############################
    # Helper methods
    ##############################

    @staticmethod
    def _check_head(src) -> None:
        """
        Check if the source exists.

        Parameters
        ----------
        src : str
            The source location.

        Returns
        -------
        None

        Raises
        ------
        HTTPError
            If an error occurs while checking the source.
        """
        r = requests.head(src, timeout=60)
        r.raise_for_status()

    def _download_file(self, url: str, dst: Path, overwrite: bool) -> str:
        """
        Method to download a file from a given url.

        Parameters
        ----------
        url : str
            The url of the file to download.
        dst : Path
            The destination of the file.
        overwrite : bool
            Whether to overwrite existing files.

        Returns
        -------
        str
            The path of the downloaded file.
        """
        self._check_head(url)
        with requests.get(url, stream=True, timeout=60) as r:
            r.raise_for_status()
            with open(dst, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return str(dst)

    def _head_extension(self, url: str, file_format: str | None = None) -> str:
        """
        Method to get the extension of a file from a given url.

        Parameters
        ----------
        url : str
            The url of the file to get the extension.
        file_format : str
            The file format to check.

        Returns
        -------
        str
            File extension.
        """
        if file_format is not None:
            return file_format
        try:
            r = requests.head(url, timeout=60)
            r.raise_for_status()
            content_type = r.headers["content-type"]
            if "text" in content_type:
                return "csv"
            else:
                raise ValueError("Content type not supported.")
        except Exception as e:
            raise e



from __future__ import annotations

import shutil
import typing
from typing import Any

from digitalhub.context.api import get_context
from digitalhub.entities._base.entity._constructors.uuid import build_uuid
from digitalhub.entities._base.material.utils import build_log_path_from_source, eval_local_source
from digitalhub.entities._commons.enums import EntityKinds, EntityTypes
from digitalhub.stores.data.api import get_store
from digitalhub.stores.readers.data.api import get_reader_by_object
from digitalhub.utils.enums import FileExtensions
from digitalhub.utils.generic_utils import slugify_string
from digitalhub.utils.types import SourcesOrListOfSources

if typing.TYPE_CHECKING:
    from digitalhub.entities.dataitem._base.entity import Dataitem


DEFAULT_EXTENSION = FileExtensions.PARQUET.value


def eval_source(
    source: SourcesOrListOfSources | None = None,
    data: Any | None = None,
    kind: str | None = None,
    name: str | None = None,
    project: str | None = None,
) -> Any:
    """
    Evaluate if source is local.

    Parameters
    ----------
    source : SourcesOrListOfSources
        Source(s).

    Returns
    -------
    None
    """
    if (source is None) == (data is None):
        raise ValueError("You must provide source or data.")

    if source is not None:
        eval_local_source(source)
        return source

    if kind == EntityKinds.DATAITEM_TABLE.value:
        ctx = get_context(project)
        pth = ctx.root / f"{slugify_string(name)}.{DEFAULT_EXTENSION}"
        reader = get_reader_by_object(data)
        reader.write_parquet(data, pth)
        return str(pth)

    raise NotImplementedError


def eval_data(
    project: str,
    kind: str,
    source: SourcesOrListOfSources,
    data: Any | None = None,
    file_format: str | None = None,
    engine: str | None = None,
) -> Any:
    """
    Evaluate data is loaded.

    Parameters
    ----------
    project : str
        Project name.
    source : str
        Source(s).
    data : Any
        Dataframe to log. Alternative to source.
    file_format : str
        Extension of the file.
    engine : str
        Engine to use.

    Returns
    -------
    None
    """
    if kind == EntityKinds.DATAITEM_TABLE.value:
        if data is None:
            return get_store(project, source).read_df(
                source,
                file_format=file_format,
                engine=engine,
            )
    return data


def process_kwargs(
    project: str,
    name: str,
    kind: str,
    source: SourcesOrListOfSources,
    data: Any | None = None,
    path: str | None = None,
    **kwargs,
) -> dict:
    """
    Process spec kwargs.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    kind : str
        Kind the object.
    source : SourcesOrListOfSources
        Source(s).
    data : Any
        Dataframe to log. Alternative to source.
    path : str
        Destination path of the entity. If not provided, it's generated.
    **kwargs : dict
        Spec parameters.

    Returns
    -------
    dict
        Kwargs updated.
    """
    if data is not None:
        if kind == EntityKinds.DATAITEM_TABLE.value:
            reader = get_reader_by_object(data)
            kwargs["schema"] = reader.get_schema(data)
    if path is None:
        uuid = build_uuid()
        kwargs["uuid"] = uuid
        kwargs["path"] = build_log_path_from_source(project, EntityTypes.DATAITEM.value, name, uuid, source)
    else:
        kwargs["path"] = path
    return kwargs


def clean_tmp_path(pth: SourcesOrListOfSources) -> None:
    """
    Clean temporary path.

    Parameters
    ----------
    pth : SourcesOrListOfSources
        Path to clean.

    Returns
    -------
    None
    """
    if isinstance(pth, list):
        for p in pth:
            shutil.rmtree(p, ignore_errors=True)
        return
    shutil.rmtree(pth, ignore_errors=True)


def post_process(obj: Dataitem, data: Any) -> Dataitem:
    """
    Post process object.

    Parameters
    ----------
    obj : Dataitem
        The object.
    data : Any
        The data.

    Returns
    -------
    Dataitem
        The object.
    """
    if obj.kind == EntityKinds.DATAITEM_TABLE.value:
        reader = get_reader_by_object(data)
        obj.status.preview = reader.get_preview(data)
        obj.save(update=True)
    return obj



from __future__ import annotations

import typing

from digitalhub.stores.client.dhcore.client import ClientDHCore
from digitalhub.stores.client.local.client import ClientLocal

if typing.TYPE_CHECKING:
    from digitalhub.stores.client._base.client import Client


class ClientBuilder:
    """
    Client builder class.

    This class is used to create two possible client instances:
    Local and DHCore.
    It saves the client instances in the class attributes using
    singleton pattern.
    """

    def __init__(self) -> None:
        self._local = None
        self._dhcore = None

    def build(self, local: bool = False) -> Client:
        """
        Method to create a client instance.

        Parameters
        ----------
        local : bool
            Whether to create a local client or not.

        Returns
        -------
        Client
            Returns the client instance.
        """
        if local:
            if self._local is None:
                self._local = ClientLocal()
            return self._local

        if self._dhcore is None:
            self._dhcore = ClientDHCore()
        return self._dhcore


client_builder = ClientBuilder()



from __future__ import annotations

import typing
from urllib.parse import urlparse

import mlflow

from digitalhub.entities.model.mlflow.models import Dataset, Signature

if typing.TYPE_CHECKING:
    from mlflow.entities import Run


def get_mlflow_run(run_id: str) -> Run:
    """
    Get MLFlow run.

    Parameters
    ----------
    run_id : str
        The id of the mlflow run.

    Returns
    -------
    Run
        The extracted run.
    """
    return mlflow.MlflowClient().get_run(run_id)


def from_mlflow_run(run_id: str) -> dict:
    """
    Extract from mlflow run spec for Digitalhub Model.

    Parameters
    ----------
    run_id : str
        The id of the mlflow run.

    Returns
    -------
    dict
        The extracted spec.
    """

    # Get MLFlow run
    run = get_mlflow_run(run_id)

    # Extract spec
    data = run.data
    parameters = data.params
    source_path = urlparse(run.info.artifact_uri).path
    model_uri = f"runs:/{run_id}/model"
    model = mlflow.pyfunc.load_model(model_uri=model_uri)
    model_config = model.model_config
    flavor = None
    for f in model.metadata.flavors:
        if f != "python_function":
            flavor = f
            break

    # Extract signature
    mlflow_signature = model.metadata.signature
    signature = Signature(
        inputs=mlflow_signature.inputs.to_json() if mlflow_signature.inputs else None,
        outputs=mlflow_signature.outputs.to_json() if mlflow_signature.outputs else None,
        params=mlflow_signature.params.to_json() if mlflow_signature.params else None,
    ).to_dict()

    # Extract datasets
    datasets = []
    if run.inputs and run.inputs.dataset_inputs:
        datasets = [
            Dataset(
                name=d.dataset.name,
                digest=d.dataset.digest,
                profile=d.dataset.profile,
                dataset_schema=d.dataset.schema,
                source=d.dataset.source,
                source_type=d.dataset.source_type,
            ).to_dict()
            for d in run.inputs.dataset_inputs
        ]

    # Create model params
    model_params = {}

    # source path
    model_params["source"] = source_path

    # common properties
    model_params["framework"] = flavor
    model_params["parameters"] = parameters

    # specific to MLFlow
    model_params["flavor"] = flavor
    model_params["model_config"] = model_config
    model_params["input_datasets"] = datasets
    model_params["signature"] = signature

    return model_params


def get_mlflow_model_metrics(run_id: str) -> dict:
    """
    Get MLFlow model metrics for a given run id.

    Parameters
    ----------
    run_id : str
        The id of the mlflow run.

    Returns
    -------
    dict
        The extracted metrics.
    """
    # Get MLFlow run
    run = get_mlflow_run(run_id)

    # Extract metrics
    return run.data.metrics



from __future__ import annotations

from digitalhub.entities.artifact.artifact.builder import ArtifactArtifactBuilder
from digitalhub.entities.dataitem.dataitem.builder import DataitemDataitemBuilder
from digitalhub.entities.dataitem.table.builder import DataitemTableBuilder
from digitalhub.entities.model.mlflow.builder import ModelModelBuilder
from digitalhub.entities.project._base.builder import ProjectProjectBuilder
from digitalhub.entities.secret._base.builder import SecretSecretBuilder

entity_builders: tuple = (
    (ProjectProjectBuilder.ENTITY_KIND, ProjectProjectBuilder),
    (SecretSecretBuilder.ENTITY_KIND, SecretSecretBuilder),
    (ArtifactArtifactBuilder.ENTITY_KIND, ArtifactArtifactBuilder),
    (DataitemDataitemBuilder.ENTITY_KIND, DataitemDataitemBuilder),
    (DataitemTableBuilder.ENTITY_KIND, DataitemTableBuilder),
    (ModelModelBuilder.ENTITY_KIND, ModelModelBuilder),
)

##############################
#  Add custom entities here
##############################


try:
    from digitalhub.entities.dataitem.iceberg.builder import DataitemIcebergBuilder

    entity_builders += ((DataitemIcebergBuilder.ENTITY_KIND, DataitemIcebergBuilder),)
except ImportError:
    ...


try:
    from digitalhub.entities.model.model.builder import ModelMlflowBuilder

    entity_builders += ((ModelMlflowBuilder.ENTITY_KIND, ModelMlflowBuilder),)
except ImportError:
    ...

try:
    from digitalhub.entities.model.sklearn.builder import ModelSklearnBuilder

    entity_builders += ((ModelSklearnBuilder.ENTITY_KIND, ModelSklearnBuilder),)
except ImportError:
    ...

try:
    from digitalhub.entities.model.huggingface.builder import ModelHuggingfaceBuilder

    entity_builders += ((ModelHuggingfaceBuilder.ENTITY_KIND, ModelHuggingfaceBuilder),)
except ImportError:
    ...



from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.model._base.builder import ModelBuilder
from digitalhub.entities.model.sklearn.entity import ModelSklearn
from digitalhub.entities.model.sklearn.spec import ModelSpecSklearn, ModelValidatorSklearn
from digitalhub.entities.model.sklearn.status import ModelStatusSklearn


class ModelSklearnBuilder(ModelBuilder):
    """
    ModelSklearn builder.
    """

    ENTITY_CLASS = ModelSklearn
    ENTITY_SPEC_CLASS = ModelSpecSklearn
    ENTITY_SPEC_VALIDATOR = ModelValidatorSklearn
    ENTITY_STATUS_CLASS = ModelStatusSklearn
    ENTITY_KIND = EntityKinds.MODEL_SKLEARN.value



from __future__ import annotations

from pydantic import BaseModel, ConfigDict

from digitalhub.entities._base._base.entity import Base


class Spec(Base):
    """
    A class representing the specifications.of an entity.
    specifications.is a collection of information about an entity
    thought to be immutable by the user.
    """

    @classmethod
    def from_dict(cls, obj: dict) -> Spec:
        """
        Return entity specifications.object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity specifications.

        Returns
        -------
        EntitySpec
            An entity specifications.object.
        """
        return cls(**obj)


class SpecValidator(BaseModel):
    """
    A class representing the parameters of an entity.
    This base class is used to define the parameters of an entity
    specifications.and is used to validate the parameters passed
    to the constructor.
    """

    model_config = ConfigDict(extra="ignore", use_enum_values=True)

    def to_dict(self) -> dict:
        return self.model_dump(by_alias=True, exclude_none=True)


class MaterialSpec(Spec):
    """
    Material specifications.class.
    """

    def __init__(self, path: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.path = path


class MaterialValidator(SpecValidator):
    """
    Material parameters class.
    """

    path: str
    """Target path to file(s)"""



from __future__ import annotations

from digitalhub.entities._base.entity.status import Status


class MaterialStatus(Status):
    """
    Material Status class.
    """

    def __init__(
        self,
        state: str,
        message: str | None = None,
        files: list[dict] | None = None,
    ) -> None:
        super().__init__(state, message)
        self.files = files if files is not None else []



from __future__ import annotations

import typing

from digitalhub.entities._base.runtime_entity.builder import EntityError, RuntimeEntityBuilder
from digitalhub.entities._base.unversioned.builder import UnversionedBuilder
from digitalhub.entities._commons.enums import EntityTypes

if typing.TYPE_CHECKING:
    from digitalhub.entities.run._base.entity import Run


class RunBuilder(UnversionedBuilder, RuntimeEntityBuilder):
    """
    Run builder.
    """

    ENTITY_TYPE = EntityTypes.RUN.value

    def build(
        self,
        project: str,
        kind: str,
        uuid: str | None = None,
        labels: list[str] | None = None,
        task: str | None = None,
        local_execution: bool = False,
        **kwargs,
    ) -> Run:
        """
        Create a new object.

        Parameters
        ----------
        project : str
            Project name.
        kind : str
            Kind the object.
        uuid : str
            ID of the object.
        labels : list[str]
            List of labels.
        task : str
            Name of the task associated with the run.
        local_execution : bool
            Flag to determine if object has local execution.
        **kwargs : dict
            Spec keyword arguments.

        Returns
        -------
        Run
            Object instance.
        """
        # Check task validity
        if task is None:
            raise EntityError("Missing task in run spec")
        self._check_kind_validity(task)

        uuid = self.build_uuid(uuid)
        metadata = self.build_metadata(
            project=project,
            name=uuid,
            labels=labels,
        )
        spec = self.build_spec(
            task=task,
            local_execution=local_execution,
            **kwargs,
        )
        status = self.build_status()
        return self.build_entity(
            project=project,
            uuid=uuid,
            kind=kind,
            metadata=metadata,
            spec=spec,
            status=status,
        )

    def _check_kind_validity(self, task: str) -> None:
        """
        Check kind validity.

        Parameters
        ----------
        task : str
            Task string.

        Returns
        -------
        None
        """
        task_kind = task.split("://")[0]
        if task_kind not in self.get_all_kinds():
            raise EntityError(f"Invalid run '{self.ENTITY_KIND}' for task kind '{task_kind}'")



from __future__ import annotations

from digitalhub.entities._base.entity.status import Status


class RunStatus(Status):
    """
    RunStatus status.
    """

    def __init__(
        self,
        state: str,
        message: str | None = None,
        transitions: list[dict] | None = None,
        k8s: dict | None = None,
        metrics: dict | None = None,
        **kwargs,
    ) -> None:
        super().__init__(state, message, transitions, k8s, **kwargs)
        self.metrics = metrics if metrics is not None else {}



from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.dataitem._base.builder import DataitemBuilder
from digitalhub.entities.dataitem.iceberg.entity import DataitemIceberg
from digitalhub.entities.dataitem.iceberg.spec import DataitemSpecIceberg, DataitemValidatorIceberg
from digitalhub.entities.dataitem.iceberg.status import DataitemStatusIceberg


class DataitemIcebergBuilder(DataitemBuilder):
    """
    DataitemIceberg builder.
    """

    ENTITY_CLASS = DataitemIceberg
    ENTITY_SPEC_CLASS = DataitemSpecIceberg
    ENTITY_SPEC_VALIDATOR = DataitemValidatorIceberg
    ENTITY_STATUS_CLASS = DataitemStatusIceberg
    ENTITY_KIND = EntityKinds.DATAITEM_ICEBERG.value



from __future__ import annotations

from digitalhub.entities._commons.enums import ApiCategories, BackendOperations
from digitalhub.stores.client._base.api_builder import ClientApiBuilder
from digitalhub.stores.client.local.enums import LocalClientVar
from digitalhub.utils.exceptions import BackendError

API_BASE = "/api/v1"
API_CONTEXT = f"{API_BASE}/-"


class ClientLocalApiBuilder(ClientApiBuilder):
    """
    This class is used to build the API for the Local client.
    """

    def build_api(self, category: str, operation: str, **kwargs) -> str:
        """
        Build the API for the client.

        Parameters
        ----------
        category : str
            API category.
        operation : str
            API operation.
        **kwargs : dict
            Additional parameters.

        Returns
        -------
        str
            API formatted.
        """
        if category == ApiCategories.BASE.value:
            return self.build_api_base(operation, **kwargs)
        return self.build_api_context(operation, **kwargs)

    def build_api_base(self, operation: str, **kwargs) -> str:
        """
        Build the base API for the client.

        Parameters
        ----------
        operation : str
            API operation.
        **kwargs : dict
            Additional parameters.

        Returns
        -------
        str
            API formatted.
        """
        entity_type = kwargs["entity_type"] + "s"
        if operation in (
            BackendOperations.CREATE.value,
            BackendOperations.LIST.value,
        ):
            return f"{API_BASE}/{entity_type}"
        elif operation in (
            BackendOperations.READ.value,
            BackendOperations.UPDATE.value,
            BackendOperations.DELETE.value,
        ):
            return f"{API_BASE}/{entity_type}/{kwargs['entity_name']}"
        raise BackendError(f"API for operation '{operation}' for entity type '{entity_type}' not implemented in Local.")

    def build_api_context(self, operation: str, **kwargs) -> str:
        """
        Build the context API for the client.
        """
        try:
            entity_type = kwargs["entity_type"] + "s"
        except KeyError:
            pass
        project = kwargs["project"]
        if operation in (
            BackendOperations.CREATE.value,
            BackendOperations.LIST.value,
        ):
            return f"{API_CONTEXT}/{project}/{entity_type}"
        elif operation in (
            BackendOperations.READ.value,
            BackendOperations.UPDATE.value,
            BackendOperations.DELETE.value,
        ):
            return f"{API_CONTEXT}/{project}/{entity_type}/{kwargs['entity_id']}"
        elif operation in (
            BackendOperations.LOGS.value,
            BackendOperations.FILES.value,
            BackendOperations.METRICS.value,
        ):
            return LocalClientVar.EMPTY.value
        raise BackendError(f"API for operation '{operation}' for entity type '{entity_type}' not implemented in Local.")



from __future__ import annotations

import base64
import importlib.util as imputil
import json
from datetime import date, datetime, time
from enum import Enum, EnumMeta
from pathlib import Path
from types import MappingProxyType
from typing import Any, Callable
from zipfile import ZipFile

import numpy as np
import requests
from slugify import slugify

from digitalhub.utils.io_utils import read_text


def get_timestamp() -> str:
    """
    Get the current timestamp timezoned.

    Returns
    -------
    str
        The current timestamp.
    """
    return datetime.now().astimezone().isoformat()


def decode_base64_string(string: str) -> str:
    """
    Decode a string from base64.

    Parameters
    ----------
    string : str
        The string to decode.

    Returns
    -------
    str
        The string decoded from base64.
    """
    return base64.b64decode(string).decode()


def encode_string(string: str) -> str:
    """
    Encode a string in base64.

    Parameters
    ----------
    string : str
        The string to encode.

    Returns
    -------
    str
        The string encoded in base64.
    """
    return base64.b64encode(string.encode()).decode()


def encode_source(path: str) -> str:
    """
    Read a file and encode in base64 the content.

    Parameters
    ----------
    path : str
        The file path to read.

    Returns
    -------
    str
        The file content encoded in base64.
    """
    return encode_string(read_text(path))


def requests_chunk_download(source: str, filename: Path) -> None:
    """
    Download a file in chunks.

    Parameters
    ----------
    source : str
        URL to download the file.
    filename : Path
        Path where to save the file.

    Returns
    -------
    None
    """
    with requests.get(source, stream=True) as r:
        r.raise_for_status()
        with filename.open("wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)


def extract_archive(path: Path, filename: Path) -> None:
    """
    Extract a zip archive.

    Parameters
    ----------
    path : Path
        Path where to extract the archive.
    filename : Path
        Path to the archive.

    Returns
    -------
    None
    """
    with ZipFile(filename, "r") as zip_file:
        zip_file.extractall(path)


class CustomJsonEncoder(json.JSONEncoder):
    """
    Custom JSON encoder to handle json dumps.
    """

    def default(self, obj: Any) -> Any:
        """
        Convert an object to json.

        Parameters
        ----------
        obj : Any
            The object to convert.

        Returns
        -------
        Any
            The object converted to json.
        """
        if isinstance(obj, (int, str, float, list, dict)):
            return obj
        elif isinstance(obj, (np.integer, np.int64)):
            return int(obj)
        elif isinstance(obj, (np.floating, np.float64)):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (datetime, date, time)):
            return obj.isoformat()
        return str(obj)


def dump_json(struct: Any) -> str:
    """
    Convert a dict to json.

    Parameters
    ----------
    struct : dict
        The dict to convert.

    Returns
    -------
    str
        The json string.
    """
    return json.dumps(struct, cls=CustomJsonEncoder)


def slugify_string(filename: str) -> str:
    """
    Sanitize a filename.

    Parameters
    ----------
    filename : str
        The filename to sanitize.

    Returns
    -------
    str
        The sanitized filename.
    """
    return slugify(filename, max_length=255)


def import_function(path: Path, handler: str) -> Callable:
    """
    Import a function from a module.

    Parameters
    ----------
    path : Path
        Path where the function source is located.
    handler : str
        Function name.

    Returns
    -------
    Callable
        Function.
    """
    spec = imputil.spec_from_file_location(path.stem, path)
    if spec is None:
        raise RuntimeError(f"Error loading function source from {str(path)}.")

    mod = imputil.module_from_spec(spec)
    if spec.loader is None:
        raise RuntimeError(f"Error getting module loader from {str(path)}.")

    spec.loader.exec_module(mod)
    func = getattr(mod, handler)
    if not callable(func):
        raise RuntimeError(f"Handler '{handler}' is not a callable.")

    return func


def list_enum(enum: EnumMeta) -> list[Any]:
    """
    Get all values of an enum.

    Parameters
    ----------
    enum : EnumMeta
        Enum class to get values from.

    Returns
    -------
    list[Any]
        List of enum values.
    """
    vals: MappingProxyType[str, Enum] = enum.__members__
    return [member.value for member in vals.values()]



from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._processors.context import context_processor
from digitalhub.utils.exceptions import EntityNotExistsError

if typing.TYPE_CHECKING:
    from digitalhub.entities.secret._base.entity import Secret


ENTITY_TYPE = EntityTypes.SECRET.value


def new_secret(
    project: str,
    name: str,
    uuid: str | None = None,
    description: str | None = None,
    labels: list[str] | None = None,
    embedded: bool = False,
    secret_value: str | None = None,
    **kwargs,
) -> Secret:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    uuid : str
        ID of the object.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object spec must be embedded in project spec.
    secret_value : str
        Value of the secret.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Secret
        Object instance.

    Examples
    --------
    >>> obj = new_secret(project="my-project",
    >>>                  name="my-secret",
    >>>                  secret_value="my-secret-value")
    """
    if secret_value is None:
        raise ValueError("secret_value must be provided.")
    obj: Secret = context_processor.create_context_entity(
        project=project,
        name=name,
        kind="secret",
        uuid=uuid,
        description=description,
        labels=labels,
        embedded=embedded,
        **kwargs,
    )
    obj.set_secret_value(value=secret_value)
    return obj


def get_secret(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    **kwargs,
) -> Secret:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Secret
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_secret("store://my-secret-key")

    Using entity name:
    >>> obj = get_secret("my-secret-name"
    >>>                  project="my-project",
    >>>                  entity_id="my-secret-id")
    """
    if not identifier.startswith("store://"):
        if project is None:
            raise ValueError("Project must be provided.")
        secrets = list_secrets(project=project, **kwargs)
        for secret in secrets:
            if secret.name == identifier:
                return secret
        else:
            raise EntityNotExistsError(f"Secret {identifier} not found.")
    return context_processor.read_context_entity(
        identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        **kwargs,
    )


def get_secret_versions(
    identifier: str,
    project: str | None = None,
    **kwargs,
) -> list[Secret]:
    """
    Get object versions from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Secret]
        List of object instances.

    Examples
    --------
    Using entity key:
    >>> objs = get_secret_versions("store://my-secret-key")

    Using entity name:
    >>> objs = get_secret_versions("my-secret-name",
    >>>                            project="my-project")
    """
    return context_processor.read_context_entity_versions(
        identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        **kwargs,
    )


def list_secrets(project: str, **kwargs) -> list[Secret]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Secret]
        List of object instances.

    Examples
    --------
    >>> objs = list_secrets(project="my-project")
    """
    return context_processor.list_context_entities(
        project=project,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )


def import_secret(file: str) -> Secret:
    """
    Import object from a YAML file and create a new object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Secret
        Object instance.

    Examples
    --------
    >>> obj = import_secret("my-secret.yaml")
    """
    return context_processor.import_context_entity(file)


def load_secret(file: str) -> Secret:
    """
    Load object from a YAML file and update an existing object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Secret
        Object instance.

    Examples
    --------
    >>> obj = load_secret("my-secret.yaml")
    """
    return context_processor.load_context_entity(file)


def update_secret(entity: Secret) -> Secret:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Secret
        Object to update.

    Returns
    -------
    Secret
        Entity updated.

    Examples
    --------
    >>> obj = update_secret(obj)
    """
    return context_processor.update_context_entity(
        project=entity.project,
        entity_type=entity.ENTITY_TYPE,
        entity_id=entity.id,
        entity_dict=entity.to_dict(),
    )


def delete_secret(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity. If True, use entity name instead of entity key as identifier.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    If delete_all_versions is False:
    >>> obj = delete_secret("store://my-secret-key")

    Otherwise:
    >>> obj = delete_secret("my-secret-name"
    >>>                     project="my-project",
    >>>                     delete_all_versions=True)
    """
    return context_processor.delete_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
        **kwargs,
    )



from __future__ import annotations

import typing

from digitalhub.entities._base.material.entity import MaterialEntity
from digitalhub.entities._commons.enums import EntityTypes

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.artifact._base.spec import ArtifactSpec
    from digitalhub.entities.artifact._base.status import ArtifactStatus


class Artifact(MaterialEntity):
    """
    A class representing a artifact.

    Artifacts are (binary) objects stored in one of the artifact
    stores of the platform, and available to every process, module
    and component as files.
    """

    ENTITY_TYPE = EntityTypes.ARTIFACT.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: ArtifactSpec,
        status: ArtifactStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        self.spec: ArtifactSpec
        self.status: ArtifactStatus



from __future__ import annotations

from enum import Enum


class DhcoreEnvVar(Enum):
    """
    Environment variables.
    """

    ENDPOINT = "DHCORE_ENDPOINT"
    ISSUER = "DHCORE_ISSUER"
    USER = "DHCORE_USER"
    PASSWORD = "DHCORE_PASSWORD"
    CLIENT_ID = "DHCORE_CLIENT_ID"
    ACCESS_TOKEN = "DHCORE_ACCESS_TOKEN"
    REFRESH_TOKEN = "DHCORE_REFRESH_TOKEN"
    WORKFLOW_IMAGE = "DHCORE_WORKFLOW_IMAGE"


class AuthType(Enum):
    """
    Authentication types.
    """

    BASIC = "basic"
    OAUTH2 = "oauth2"



from __future__ import annotations

from enum import Enum


class EntityTypes(Enum):
    """
    Entity types.
    """

    PROJECT = "project"
    ARTIFACT = "artifact"
    DATAITEM = "dataitem"
    MODEL = "model"
    SECRET = "secret"
    FUNCTION = "function"
    WORKFLOW = "workflow"
    TASK = "task"
    RUN = "run"


class Relationship(Enum):
    """
    Relationship enumeration.
    """

    PRODUCEDBY = "produced_by"
    CONSUMES = "consumes"
    RUN_OF = "run_of"


class State(Enum):
    """
    State enumeration.
    """

    BUILT = "BUILT"
    CANCELLED = "CANCELLED"
    COMPLETED = "COMPLETED"
    CREATED = "CREATED"
    CREATING = "CREATING"
    DELETED = "DELETED"
    DELETING = "DELETING"
    ERROR = "ERROR"
    FSM_ERROR = "FSM_ERROR"
    IDLE = "IDLE"
    NONE = "NONE"
    ONLINE = "ONLINE"
    PENDING = "PENDING"
    READY = "READY"
    RESUME = "RESUME"
    RUN_ERROR = "RUN_ERROR"
    RUNNING = "RUNNING"
    STOP = "STOP"
    STOPPED = "STOPPED"
    SUCCESS = "SUCCESS"
    UNKNOWN = "UNKNOWN"


class ApiCategories(Enum):
    """
    Entity categories.
    """

    BASE = "base"
    CONTEXT = "context"


class BackendOperations(Enum):
    """
    Backend operations.
    """

    CREATE = "create"
    READ = "read"
    READ_ALL_VERSIONS = "read_all_versions"
    UPDATE = "update"
    DELETE = "delete"
    LIST = "list"
    LIST_FIRST = "list_first"
    STOP = "stop"
    RESUME = "resume"
    DATA = "data"
    FILES = "files"
    LOGS = "logs"
    SEARCH = "search"
    SHARE = "share"
    METRICS = "metrics"


class EntityKinds(Enum):
    """
    Entity kinds.
    """

    PROJECT_PROJECT = "project"
    ARTIFACT_ARTIFACT = "artifact"
    DATAITEM_DATAITEM = "dataitem"
    DATAITEM_TABLE = "table"
    DATAITEM_ICEBERG = "iceberg"
    MODEL_MODEL = "model"
    MODEL_MLFLOW = "mlflow"
    MODEL_HUGGINGFACE = "huggingface"
    MODEL_SKLEARN = "sklearn"
    SECRET_SECRET = "secret"



from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.model._base.builder import ModelBuilder
from digitalhub.entities.model.mlflow.entity import ModelMlflow
from digitalhub.entities.model.mlflow.spec import ModelSpecMlflow, ModelValidatorMlflow
from digitalhub.entities.model.mlflow.status import ModelStatusMlflow


class ModelModelBuilder(ModelBuilder):
    """
    ModelMlflow builder.
    """

    ENTITY_CLASS = ModelMlflow
    ENTITY_SPEC_CLASS = ModelSpecMlflow
    ENTITY_SPEC_VALIDATOR = ModelValidatorMlflow
    ENTITY_STATUS_CLASS = ModelStatusMlflow
    ENTITY_KIND = EntityKinds.MODEL_MLFLOW.value



from __future__ import annotations

from digitalhub.utils.exceptions import EntityError


class RuntimeEntityBuilder:
    """
    RuntimeEntity builder.
    """

    EXECUTABLE_KIND: str = None
    TASKS_KINDS: dict = None
    RUN_KIND: str = None

    def __init__(self) -> None:
        if self.EXECUTABLE_KIND is None:
            raise EntityError("EXECUTABLE_KIND must be set")
        if self.TASKS_KINDS is None:
            raise EntityError("TASKS_KINDS must be set")
        if self.RUN_KIND is None:
            raise EntityError("RUN_KIND must be set")

    def get_action_from_task_kind(self, task_kind: str) -> str:
        """
        Get action from task kind.

        Parameters
        ----------
        task_kind : str
            Task kind.

        Returns
        -------
        str
            Action.
        """
        for task in self.TASKS_KINDS:
            if task["kind"] == task_kind:
                return task["action"]
        msg = f"Task kind {task_kind} not allowed."
        raise EntityError(msg)

    def get_task_kind_from_action(self, action: str) -> list[str]:
        """
        Get task kinds from action.

        Parameters
        ----------
        action : str
            Action.

        Returns
        -------
        list[str]
            Task kinds.
        """
        for task in self.TASKS_KINDS:
            if task["action"] == action:
                return task["kind"]
        msg = f"Action {action} not allowed."
        raise EntityError(msg)

    def get_run_kind(self) -> str:
        """
        Get run kind.

        Returns
        -------
        str
            Run kind.
        """
        return self.RUN_KIND

    def get_executable_kind(self) -> str:
        """
        Get executable kind.

        Returns
        -------
        str
            Executable kind.
        """
        return self.EXECUTABLE_KIND

    def get_all_kinds(self) -> list[str]:
        """
        Get all kinds.

        Returns
        -------
        list[str]
            All kinds.
        """
        task_kinds = [i["kind"] for i in self.TASKS_KINDS]
        return [self.EXECUTABLE_KIND, self.RUN_KIND, *task_kinds]

    def get_all_actions(self) -> list[str]:
        """
        Get all actions.

        Returns
        -------
        list[str]
            All actions.
        """
        return [i["action"] for i in self.TASKS_KINDS]



from __future__ import annotations

from pydantic import Field

from digitalhub.entities.dataitem._base.spec import DataitemSpec, DataitemValidator
from digitalhub.entities.dataitem.table.models import TableSchema


class DataitemSpecTable(DataitemSpec):
    """
    DataitemSpecTable specifications.
    """

    def __init__(self, path: str, schema: dict | None = None) -> None:
        super().__init__(path)
        self.schema = schema


class DataitemValidatorTable(DataitemValidator):
    """
    DataitemValidatorTable validator.
    """

    schema_: TableSchema = Field(default=None, alias="schema")
    """The schema of the dataitem in table schema format."""



from __future__ import annotations

import typing
from typing import Any

from requests import request
from requests.exceptions import JSONDecodeError

from digitalhub.stores.client._base.client import Client
from digitalhub.stores.client.dhcore.api_builder import ClientDHCoreApiBuilder
from digitalhub.stores.client.dhcore.configurator import ClientDHCoreConfigurator
from digitalhub.stores.client.dhcore.error_parser import ErrorParser
from digitalhub.stores.client.dhcore.key_builder import ClientDHCoreKeyBuilder
from digitalhub.stores.client.dhcore.params_builder import ClientDHCoreParametersBuilder
from digitalhub.utils.exceptions import BackendError
from digitalhub.utils.generic_utils import dump_json

if typing.TYPE_CHECKING:
    from requests import Response


class ClientDHCore(Client):
    """
    DHCore client.

    The DHCore client is used to communicate with the Digitalhub Core
    backendAPI via REST. The client supports basic authentication and
    OAuth2 token authentication with token refresh.
    At creation, the client tries to get the endpoint and authentication
    parameters from the .dhcore file and the environment variables. In
    case the user incours into an authentication/endpoint error during
    the client creation, the user has the possibility to update the
    correct parameters using the `set_dhcore_env` function. If the DHCore
    client is already initialized, this function will override the
    configuration, otherwise it simply set the environment variables.
    """

    def __init__(self) -> None:
        super().__init__()

        # API builder
        self._api_builder = ClientDHCoreApiBuilder()

        # Key builder
        self._key_builder = ClientDHCoreKeyBuilder()

        # Parameters builder
        self._params_builder = ClientDHCoreParametersBuilder()

        # Error parser
        self._error_parser = ErrorParser()

        # Client Configurator
        self._configurator = ClientDHCoreConfigurator()
        self._configurator.configure()

    ##############################
    # CRUD methods
    ##############################

    def create_object(self, api: str, obj: Any, **kwargs) -> dict:
        """
        Create an object in DHCore.

        Parameters
        ----------
        api : str
            Create API.
        obj : Any
            Object to create.
        **kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        dict
            Response object.
        """
        if "headers" not in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["Content-Type"] = "application/json"
        kwargs["data"] = dump_json(obj)
        return self._prepare_call("POST", api, **kwargs)

    def read_object(self, api: str, **kwargs) -> dict:
        """
        Get an object from DHCore.

        Parameters
        ----------
        api : str
            Read API.
        **kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        dict
            Response object.
        """
        return self._prepare_call("GET", api, **kwargs)

    def update_object(self, api: str, obj: Any, **kwargs) -> dict:
        """
        Update an object in DHCore.

        Parameters
        ----------
        api : str
            Update API.
        obj : dict
            Object to update.
        **kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        dict
            Response object.
        """
        if "headers" not in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["Content-Type"] = "application/json"
        kwargs["data"] = dump_json(obj)
        return self._prepare_call("PUT", api, **kwargs)

    def delete_object(self, api: str, **kwargs) -> dict:
        """
        Delete an object from DHCore.

        Parameters
        ----------
        api : str
            Delete API.
        **kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        dict
            Response object.
        """
        resp = self._prepare_call("DELETE", api, **kwargs)
        if isinstance(resp, bool):
            resp = {"deleted": resp}
        return resp

    def list_objects(self, api: str, **kwargs) -> list[dict]:
        """
        List objects from DHCore.

        Parameters
        ----------
        api : str
            List API.
        **kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        list[dict]
            Response objects.
        """
        if kwargs is None:
            kwargs = {}

        if "params" not in kwargs:
            kwargs["params"] = {}

        start_page = 0
        if "page" not in kwargs["params"]:
            kwargs["params"]["page"] = start_page

        objects = []
        while True:
            resp = self._prepare_call("GET", api, **kwargs)
            contents = resp["content"]
            total_pages = resp["totalPages"]
            if not contents or kwargs["params"]["page"] >= total_pages:
                break
            objects.extend(contents)
            kwargs["params"]["page"] += 1

        return objects

    def list_first_object(self, api: str, **kwargs) -> dict:
        """
        List first objects.

        Parameters
        ----------
        api : str
            The api to list the objects with.
        **kwargs : dict
            Keyword arguments passed to the request.

        Returns
        -------
        dict
            The list of objects.
        """
        try:
            return self.list_objects(api, **kwargs)[0]
        except IndexError:
            raise BackendError("No object found.")

    def search_objects(self, api: str, **kwargs) -> list[dict]:
        """
        Search objects from DHCore.

        Parameters
        ----------
        api : str
            Search API.
        **kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        list[dict]
            Response objects.
        """
        if kwargs is None:
            kwargs = {}

        if "params" not in kwargs:
            kwargs["params"] = {}

        start_page = 0
        if "page" not in kwargs["params"]:
            kwargs["params"]["page"] = start_page

        if "size" not in kwargs["params"]:
            kwargs["params"]["size"] = 10

        # Add sorting
        if "sort" not in kwargs["params"]:
            kwargs["params"]["sort"] = "metadata.updated,DESC"

        objects_with_highlights: list[dict] = []
        while True:
            resp = self._prepare_call("GET", api, **kwargs)
            contents = resp["content"]
            total_pages = resp["totalPages"]
            if not contents or kwargs["params"]["page"] >= total_pages:
                break
            objects_with_highlights.extend(contents)
            kwargs["params"]["page"] += 1

        objects = []
        for obj in objects_with_highlights:
            obj.pop("highlights", None)
            objects.append(obj)

        return objects

    ##############################
    # Call methods
    ##############################

    def _prepare_call(self, call_type: str, api: str, **kwargs) -> dict:
        """
        Prepare a call to the DHCore API.

        Parameters
        ----------
        call_type : str
            The type of call to prepare.
        api : str
            The api to call.
        **kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        dict
            Response object.
        """
        self._configurator.check_config()
        if kwargs is None:
            kwargs = {}
        url = self._configurator.build_url(api)
        kwargs = self._configurator.set_request_auth(kwargs)
        return self._make_call(call_type, url, **kwargs)

    def _make_call(self, call_type: str, url: str, refresh_token: bool = True, **kwargs) -> dict:
        """
        Make a call to the DHCore API.

        Parameters
        ----------
        call_type : str
            The type of call to make.
        url : str
            The URL to call.
        **kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        dict
            Response object.
        """
        # Call the API
        response = request(call_type, url, timeout=60, **kwargs)

        # Evaluate DHCore API version
        self._configurator.check_core_version(response)

        # Handle token refresh
        if response.status_code in [401] and refresh_token and self._configurator.oauth2_auth():
            self._configurator.get_new_access_token()
            kwargs = self._configurator.set_request_auth(kwargs)
            return self._make_call(call_type, url, refresh_token=False, **kwargs)

        self._error_parser.parse(response)
        return self._dictify_response(response)

    def _dictify_response(self, response: Response) -> dict:
        """
        Return dict from response.

        Parameters
        ----------
        response : Response
            The response object.

        Returns
        -------
        dict
            The parsed response object.
        """
        try:
            return response.json()
        except JSONDecodeError:
            if response.text == "":
                return {}
            raise BackendError("Backend response could not be parsed.")

    ##############################
    # Interface methods
    ##############################

    @staticmethod
    def is_local() -> bool:
        """
        Declare if Client is local.

        Returns
        -------
        bool
            False
        """
        return False



from __future__ import annotations

import typing

from digitalhub.entities.model._base.entity import Model

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.model.model.spec import ModelSpecModel
    from digitalhub.entities.model.model.status import ModelStatusModel


class ModelModel(Model):
    """
    ModelModel class.
    """

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: ModelSpecModel,
        status: ModelStatusModel,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        self.spec: ModelSpecModel
        self.status: ModelStatusModel



import json
import os
from glob import glob
from pathlib import Path

import pytest
from digitalhub.factory.factory import factory
from jsonschema import validate

entities_path = "test/local/instances/entities"
schemas_path = "test/local/instances/schemas"

# Build dict: kind -> path to schema file
schemas = {}
for path_to_schema in glob(f"{schemas_path}/**/*.json", recursive=True):
    kind = Path(path_to_schema).stem
    schemas[kind] = path_to_schema

# Build dict: name of file to validate -> full path to file
entity_paths = {}
for path_to_file in glob(f"{entities_path}/**/*.json", recursive=True):
    file_name = os.path.basename(path_to_file)

    # If a file in a nested directory causes a name collision, use its full path as name
    if file_name in entity_paths:
        file_name = path_to_file

    entity_paths[file_name] = path_to_file


# Build object from JSON file using factory
def build_obj(entity_file_path):
    with open(entity_file_path) as f:
        entity = json.load(f)

    kind = entity["kind"]
    spec = entity["spec"]

    built = factory.build_spec(kind, **spec)
    return built.to_dict(), kind


# Validate built object against its kind's schema
def is_valid(built, kind):
    with open(schemas[kind]) as schema_file:
        schema = json.load(schema_file)

    validate(instance=built, schema=schema)
    return True


# Tests that each JSON file contained in the specified path can successfully be
# used to generate an object through the factory, and that each generated object,
# when exported to dict, validates (through jsonschema) against its kind's schema.
class TestValidate:
    @pytest.mark.parametrize("file_name", list(entity_paths.keys()))
    def test_validate(self, file_name):
        built, kind = build_obj(f"{entity_paths[file_name]}")
        assert is_valid(built, kind)



from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator


class WorkflowSpec(Spec):
    """
    WorkflowSpec specifications.
    """


class WorkflowValidator(SpecValidator):
    """
    WorkflowValidator validator.
    """



from __future__ import annotations


def build_task_actions(kind_action_list: list[tuple[str, str]]) -> list[dict[str, str]]:
    """
    Build task actions.

    Parameters
    ----------
    kind_action_list : list[tuple[str, str]]
        List of kind-action couples.

    Returns
    -------
    list[dict[str, str]]
        Returns the task actions.
    """
    return [{"kind": kind, "action": action} for (kind, action) in kind_action_list]



from __future__ import annotations

import typing

from digitalhub.entities._base.versioned.entity import VersionedEntity
from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._processors.context import context_processor
from digitalhub.entities.run.crud import delete_run, get_run, list_runs
from digitalhub.entities.task.crud import delete_task
from digitalhub.factory.api import build_entity_from_dict, build_entity_from_params
from digitalhub.utils.exceptions import EntityAlreadyExistsError, EntityError

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities._base.entity.spec import Spec
    from digitalhub.entities._base.entity.status import Status
    from digitalhub.entities.run._base.entity import Run
    from digitalhub.entities.task._base.entity import Task


class ExecutableEntity(VersionedEntity):
    """
    A class representing an entity that can be executed.
    """

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: Spec,
        status: Status,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        # Initialize tasks
        self._tasks: dict[str, Task] = {}

    ##############################
    #  Helpers
    ##############################

    def _get_executable_string(self) -> str:
        """
        Get executable string.

        Returns
        -------
        str
            Executable string.
        """
        return f"{self.kind}://{self.project}/{self.name}:{self.id}"

    ##############################
    #  Tasks
    ##############################

    def _get_or_create_task(self, kind: str) -> Task:
        """
        Get or create task.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        Task
            Task.
        """
        if self._tasks.get(kind) is None:
            try:
                self._tasks[kind] = self.get_task(kind)
            except EntityError:
                self._tasks[kind] = self.new_task(kind)
        return self._tasks[kind]

    def import_tasks(self, tasks: list[dict]) -> None:
        """
        Import tasks from yaml.

        Parameters
        ----------
        tasks : list[dict]
            List of tasks to import.

        Returns
        -------
        None
        """
        # Loop over tasks list, in the case where the function
        # is imported from local file.
        for task in tasks:
            # If task is not a dictionary, skip it
            if not isinstance(task, dict):
                continue

            # Create a new object from dictionary.
            # the form in which tasks are stored in function
            # status
            task_obj = build_entity_from_dict(task)

            # Try to save it in backend to been able to use
            # it for launching runs. In fact, tasks must be
            # persisted in backend to be able to launch runs.
            # Ignore if task already exists
            try:
                task_obj.save()
            except EntityAlreadyExistsError:
                pass

            # Set task if function is the same. Overwrite
            # status task dict with the new task object
            if task_obj.spec.function == self._get_executable_string():
                self._tasks[task_obj.kind] = task_obj

    def new_task(self, task_kind: str, **kwargs) -> Task:
        """
        Create new task. If the task already exists, update it.

        Parameters
        ----------
        task_kind : str
            Kind the object.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Task
            New task.
        """
        self._raise_if_exists(task_kind)

        if kwargs is None:
            kwargs = {}

        # Override kwargs
        kwargs["project"] = self.project
        kwargs[self.ENTITY_TYPE] = self._get_executable_string()
        kwargs["kind"] = task_kind

        # Create object instance
        task = build_entity_from_params(**kwargs)
        task.save()

        self._tasks[task_kind] = task
        return task

    def get_task(self, kind: str) -> Task:
        """
        Get task.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        Task
            Task.

        Raises
        ------
        EntityError
            If task is not created.
        """
        try:
            return self._tasks[kind]
        except KeyError:
            resp = self._get_task_from_backend(kind)
            if not resp:
                raise EntityError(f"Task {kind} is not created")
            self._tasks[kind] = resp[0]
            return self._tasks[kind]

    def update_task(self, kind: str, **kwargs) -> Task:
        """
        Update task.

        Parameters
        ----------
        kind : str
            Kind the object.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Task
            Task.
        """
        self._raise_if_not_exists(kind)

        if kwargs is None:
            kwargs = {}

        # Update kwargs
        kwargs["project"] = self.project
        kwargs["kind"] = kind
        kwargs[self.ENTITY_TYPE] = self._get_executable_string()
        kwargs["uuid"] = self._tasks[kind].id

        # Update task
        task = build_entity_from_params(**kwargs)
        task.save(update=True)
        self._tasks[kind] = task
        return task

    def delete_task(self, kind: str, cascade: bool = True) -> dict:
        """
        Delete task.

        Parameters
        ----------
        kind : str
            Kind the object.
        cascade : bool
            Flag to determine if cascade deletion must be performed.

        Returns
        -------
        dict
            Response from backend.
        """
        resp = delete_task(self._tasks[kind].key, cascade=cascade)
        self._tasks.pop(kind, None)
        return resp

    def _get_task_from_backend(self, kind: str) -> list:
        """
        List tasks from backend filtered by function and kind.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        list
            Response from backend.
        """
        params = {self.ENTITY_TYPE: self._get_executable_string(), "kind": kind}
        return context_processor.list_context_entities(self.project, EntityTypes.TASK.value, params=params)

    def _check_task_in_backend(self, kind: str) -> bool:
        """
        Check if task exists in backend.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        bool
            Flag to determine if task exists in backend.
        """
        resp = self._get_task_from_backend(kind)
        if not resp:
            return False
        return True

    def _raise_if_exists(self, kind: str) -> None:
        """
        Raise error if task is created.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task already exists.
        """
        if self._check_task_in_backend(kind):
            raise EntityError(f"Task '{kind}' already exists.")

    def _raise_if_not_exists(self, kind: str) -> None:
        """
        Raise error if task is not created.

        Parameters
        ----------
        kind : str
            Kind the object.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task does not exist.
        """
        if self._tasks.get(kind) is None:
            raise EntityError(f"Task '{kind}' does not exist.")

    ##############################
    #  Runs
    ##############################

    def get_run(
        self,
        identifier: str,
        **kwargs,
    ) -> Run:
        """
        Get object from backend.

        Parameters
        ----------
        identifier : str
            Entity key (store://...) or entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        Run
            Object instance.

        Examples
        --------
        Using entity key:
        >>> obj = executable.get_run("store://my-secret-key")

        Using entity ID:
        >>> obj = executable.get_run("123")
        """
        obj = get_run(
            identifier=identifier,
            project=self.project,
            **kwargs,
        )
        self.refresh()
        return obj

    def list_runs(self, **kwargs) -> list[Run]:
        """
        List all runs from backend.

        Parameters
        ----------
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        list[Run]
            List of object instances.

        Examples
        --------
        >>> objs = executable.list_runs()
        """
        if kwargs is None:
            kwargs = {}
        kwargs["params"] = {self.ENTITY_TYPE: self._get_executable_string()}
        return list_runs(self.project, **kwargs)

    def delete_run(
        self,
        identifier: str,
        **kwargs,
    ) -> None:
        """
        Delete run from backend.

        Parameters
        ----------
        identifier : str
            Entity key (store://...) or entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Response from backend.

        Examples
        --------
        >>> executable.delete_run("store://my-run-key")

        """
        delete_run(
            identifier=identifier,
            project=self.project,
            **kwargs,
        )
